"""Base inventory tree management
"""
import xml.etree.ElementTree as Et

from ...core import fix_name
from ...core.inventory.indexes import (CatalogIndex, ComponentIndex,
                                       DatasetIndex, ParameterIndex,
                                       ArgumentListIndex, ArgumentIndex,
                                       TemplatedParameterIndex, SpeasyIndex,
                                       TimetableIndex)


def to_xmlid(index_or_str) -> str:
    if type(index_or_str) is str:
        return index_or_str
    if type(index_or_str) is dict and "xmlid" in index_or_str:
        return index_or_str['xmlid']
    if hasattr(index_or_str, 'xmlid'):
        return index_or_str.xmlid
    else:
        raise TypeError(f"given parameter {index_or_str} of type {type(index_or_str)} is not a compatible index")


class ImpexXMLParser:

    @staticmethod
    def fix_names(**kwargs):
        return {fix_name(key): value for key, value in kwargs.items()}

    @staticmethod
    def fix_xmlid(**kwargs):
        def clean(key):
            if '}id' in key and key.startswith('{'):
                return 'xmlid'
            return key

        return {clean(key): value for key, value in kwargs.items()}

    @staticmethod
    def index_ctor_args(node, provider_name, is_public: bool = True):
        meta = ImpexXMLParser.fix_names(**ImpexXMLParser.fix_xmlid(**node.attrib))
        meta['is_public'] = is_public
        if 'dataStart' in meta:
            meta['start_date'] = meta.pop('dataStart')
        if 'dataStop' in meta:
            meta['stop_date'] = meta.pop('dataStop')
        uid = meta.get('xmlid', node.tag)
        return {"name": meta.get('name', node.tag), "provider": provider_name, 'uid': uid, "meta": meta}

    @staticmethod
    def make_any_node(parent, node, provider_name, ctor, name_key='xmlid', is_public: bool = True):
        new = ctor(**ImpexXMLParser.index_ctor_args(node, provider_name, is_public=is_public))
        name = fix_name(new.__dict__.get(name_key, node.tag))
        parent.__dict__[name] = new
        return new

    @staticmethod
    def make_dataset_node(parent, node, provider_name, name_key, is_public: bool = True):
        ds = ImpexXMLParser.make_any_node(parent, node, provider_name, DatasetIndex, name_key=name_key,
                                          is_public=is_public)
        return ds

    @staticmethod
    def make_parameter_node(parent, node, provider_name, name_key, is_public: bool = True):
        if arguments:=node.find('.//arguments'):
            arguments.set('name', '__spz_arguments__')
            param = ImpexXMLParser.make_any_node(parent, node, provider_name, TemplatedParameterIndex, name_key=name_key,
                                                 is_public=is_public)
        else:
            param = ImpexXMLParser.make_any_node(parent, node, provider_name, ParameterIndex, name_key=name_key,
                                                 is_public=is_public)
        if isinstance(parent, DatasetIndex):
            param.start_date = parent.start_date
            param.stop_date = parent.stop_date
            param.dataset = parent.spz_uid()
        return param

    @staticmethod
    def make_user_parameter_node(parent, node, provider_name, name_key, is_public: bool = True):
        # It seems that AMDA prevents users from using incompatible names here
        param = ImpexXMLParser.make_any_node(parent, node, provider_name, ParameterIndex, name_key=name_key,
                                             is_public=is_public)
        return param

    @staticmethod
    def make_component_node(parent, node, provider_name, name_key, is_public: bool = True):
        component = ImpexXMLParser.make_any_node(parent, node, provider_name, ComponentIndex, name_key=name_key,
                                                 is_public=is_public)
        if isinstance(parent, ParameterIndex):
            component.start_date = parent.start_date
            component.stop_date = parent.stop_date
            component.dataset = parent.dataset
            component.parameter = parent.spz_uid()
        return component

    @staticmethod
    def make_timetable_node(parent, node, provider_name, name_key, is_public: bool = True):
        tt = ImpexXMLParser.make_any_node(parent, node, provider_name, TimetableIndex, name_key=name_key,
                                          is_public=is_public)
        return tt

    @staticmethod
    def make_catalog_node(parent, node, provider_name, name_key, is_public: bool = True):
        cat = ImpexXMLParser.make_any_node(parent, node, provider_name, CatalogIndex, name_key=name_key,
                                           is_public=is_public)
        return cat

    @staticmethod
    def make_path_node(parent, node, provider_name, name_key, is_public: bool = True):
        return ImpexXMLParser.make_any_node(parent, node, provider_name, SpeasyIndex, name_key=name_key,
                                            is_public=is_public)

    @staticmethod
    def parse_template_arguments(parent, node, provider_name, name_key, is_public: bool = True):
        return ImpexXMLParser.make_any_node(parent, node, provider_name, ArgumentListIndex, name_key=name_key,
                                            is_public=is_public)

    @staticmethod
    def parse_template_argument(parent, node, provider_name, name_key, is_public: bool = True):
        if node.get('type') == 'list':
            choices = []
            for item in node.findall('.//item'):
                choices.append((item.get('name'), item.get('key')))
                node.remove(item)
            node.set('choices', choices)
        elif node.get('type') == 'generated-list':
            node.set('type', 'list')
            choices = []
            for k in range(int(node.get('minkey')), int(node.get('maxkey'))):
                choices.append((node.get('nametpl').replace('##key##', str(k), 1), str(k)))
            node.set('choices', choices)

        return ImpexXMLParser.make_any_node(parent, node, provider_name, ArgumentIndex, name_key=name_key,
                                            is_public=is_public)


    @staticmethod
    def parse(xml, provider_name, name_mapping=None, is_public: bool = True):
        handlers = {
            'mission': ImpexXMLParser.make_path_node,
            'observatory': ImpexXMLParser.make_path_node,
            'datasetGroup': ImpexXMLParser.make_path_node,
            'instrument': ImpexXMLParser.make_path_node,
            'dataset': ImpexXMLParser.make_dataset_node,
            'parameter': ImpexXMLParser.make_parameter_node,
            'component': ImpexXMLParser.make_component_node,
            'timeTable': ImpexXMLParser.make_timetable_node,
            'timetab': ImpexXMLParser.make_timetable_node,
            'catalog': ImpexXMLParser.make_catalog_node,
            'param': ImpexXMLParser.make_user_parameter_node,
            'arguments': ImpexXMLParser.parse_template_arguments,
            'argument': ImpexXMLParser.parse_template_argument
        }

        def _recursive_parser(parent, node, is_node_public):
            if name_mapping and node.tag in name_mapping:
                name_key = name_mapping[node.tag]
            elif node.tag not in handlers.keys():
                name_key = ''
            else:
                name_key = 'name'
            new = handlers.get(node.tag, ImpexXMLParser.make_path_node)(parent, node, provider_name, name_key,
                                                                        is_node_public)
            for subnode in node:
                _recursive_parser(new, subnode, is_node_public)

        root = SpeasyIndex("root", provider_name, f"{provider_name}_root_node")
        if xml is not None:
            tree = Et.fromstring(xml)
            _recursive_parser(root, tree, is_node_public=is_public)

        return root



import logging
from datetime import datetime, timedelta
from enum import Enum
from typing import Dict, Optional, List, Union
from types import SimpleNamespace
import warnings
from copy import deepcopy

import numpy as np

from ..datetime_range import DateTimeRange
from ...core.dataprovider import (DataProvider)

from ...core import make_utc_datetime

from ...core.inventory.indexes import (ComponentIndex, DatasetIndex, ParameterIndex, SpeasyIndex,
                                       TimetableIndex, CatalogIndex, TemplatedParameterIndex, AnyProductIndex)
from ...core.codecs import get_codec
from ...core.proxy import MINIMUM_REQUIRED_PROXY_VERSION
from ...products.variable import SpeasyVariable, merge, DataContainer
from ...products.catalog import Catalog
from ...products.dataset import Dataset
from ...products.timetable import TimeTable
from ...products import MaybeAnyProduct

from ...inventories import flat_inventories

from .parser import ImpexXMLParser, to_xmlid
from .client import ImpexClient, ImpexEndpoint
from .utils import load_catalog, load_timetable, is_private, is_public
from .exceptions import MissingCredentials

log = logging.getLogger(__name__)


class ImpexProductType(Enum):
    """Enumeration of the type of products available in an impex webservice.
    """
    UNKNOWN = 0
    DATASET = 1
    PARAMETER = 2
    COMPONENT = 3
    TIMETABLE = 4
    CATALOG = 5


class ImpexProvider(DataProvider):
    """ImpexProvider class. This class is the main interface to interact with the Impex data providers such as AMDA or CLWeb.
    """

    def __init__(self, provider_name: str, server_url: str, max_chunk_size_days: int = 10, capabilities: List = None,
                 username: str = "", password: str = "", name_mapping: Dict = None, output_format: str = 'CDF',
                 min_proxy_version=MINIMUM_REQUIRED_PROXY_VERSION):
        self.provider_name = provider_name
        self.server_url = server_url
        self.client = ImpexClient(capabilities=capabilities, server_url=server_url,
                                  username=username, password=password, output_format=output_format)
        if not self.client.is_alive():
            warnings.warn(f"The data provider {provider_name} appears to be under maintenance")
        self.max_chunk_size_days = max_chunk_size_days
        self.name_mapping = name_mapping
        self._cdf_codec = get_codec('application/x-cdf')
        DataProvider.__init__(self, provider_name=provider_name, min_proxy_version=min_proxy_version)

    def reset_credentials(self, username: str = "", password: str = ""):
        """Reset user credentials and update the inventory by replacing the information contained in the configuration

        Parameters
        ----------
        username: Optional[str]
            username in the related service
        password: Optional[str]
            user password in the related service

        """
        self.client.reset_credentials(username=username, password=password)
        self.update_inventory()

    def credential_are_valid(self):
        """Tels if credential are valid. It only checks they are not empty. It does not check if they are correct.

        Returns
        -------
        bool
            True if credentials are valid, False otherwise.

        """
        return self.client.credential_are_valid()

    def build_inventory(self, root: SpeasyIndex):
        """Build public inventory from the specified Impex data provider.

        Parameters
        ----------
        root: SpeasyIndex
            root index in which to add all the indexes in the data provider's inventory

        Returns
        -------
        SpeasyIndex
            root index populated with public inventory

        """
        obs_data_tree = ImpexXMLParser.parse(self._get_obs_data_tree(), self.provider_name, self.name_mapping)
        root.Parameters = SpeasyIndex(name='Parameters', provider=self.provider_name, uid='Parameters',
                                      meta=obs_data_tree.dataRoot.dataCenter.__dict__)

        if self.client.is_capable(ImpexEndpoint.GETTT):
            root.TimeTables = SpeasyIndex(name='TimeTables', provider=self.provider_name, uid='TimeTables')
            public_tt = ImpexXMLParser.parse(self._get_timetables_tree(), self.provider_name, self.name_mapping)
            if hasattr(public_tt, 'ws'):
                # CLWeb case
                shared_root = public_tt.ws.timetabList
            else:
                # AMDA case
                shared_root = public_tt.timeTableList
            root.TimeTables.SharedTimeTables = SpeasyIndex(name='SharedTimeTables', provider=self.provider_name,
                                                           uid='SharedTimeTables',
                                                           meta=shared_root.__dict__)

        if self.client.is_capable(ImpexEndpoint.GETCAT):
            root.Catalogs = SpeasyIndex(name='Catalogs', provider=self.provider_name, uid='Catalogs')
            public_cat = ImpexXMLParser.parse(self._get_catalogs_tree(), self.provider_name, self.name_mapping)
            root.Catalogs.SharedCatalogs = SpeasyIndex(name='SharedCatalogs', provider=self.provider_name,
                                                       uid='SharedCatalogs',
                                                       meta=public_cat.catalogList.__dict__)

        return root

    def build_private_inventory(self, root: SpeasyIndex):
        """Build private inventory, requires user authentication.

        Parameters
        ----------
        root: SpeasyIndex
            root index in which to add all the indexes in the data provider's inventory

        Returns
        -------
        SpeasyIndex
            root index populated with user private inventory

        """
        if self.client.credential_are_valid():
            if self.client.is_capable(ImpexEndpoint.GETTT):
                user_tt = ImpexXMLParser.parse(self._get_user_timetables_tree(),
                                               self.provider_name, self.name_mapping, is_public=False)
                if hasattr(user_tt, 'ws'):
                    # CLWeb case
                    public_root = user_tt.ws.timetabList
                else:
                    # AMDA case
                    public_root = user_tt.timetabList
                root.TimeTables.MyTimeTables = SpeasyIndex(name='MyTimeTables', provider=self.provider_name,
                                                           uid='MyTimeTables', meta=public_root.__dict__)

            if self.client.is_capable(ImpexEndpoint.GETCAT):
                user_cat = ImpexXMLParser.parse(self._get_user_catalogs_tree(), self.provider_name,
                                                self.name_mapping, is_public=False)
                root.Catalogs.MyCatalogs = SpeasyIndex(name='MyCatalogs', provider=self.provider_name, uid='MyCatalogs',
                                                       meta=user_cat.catalogList.__dict__)

            if self.client.is_capable(ImpexEndpoint.LISTPARAM):
                user_param = ImpexXMLParser.parse(self._get_derived_parameter_tree(),
                                                  self.provider_name, self.name_mapping, is_public=False)
                root.DerivedParameters = SpeasyIndex(name='DerivedParameters', provider=self.provider_name,
                                                     uid='DerivedParameters', meta=user_param.ws.paramList.__dict__)
        return root

    def parameter_range(self, parameter_id: str or ParameterIndex) -> Optional[DateTimeRange]:
        """Get parameter time range as defined in the inventory.

        Parameters
        ----------
        parameter_id: str or ParameterIndex
            parameter id

        Returns
        -------
        Optional[DateTimeRange]
            Data time range

        Examples
        --------
        >>> import speasy as spz
        >>> spz.amda.parameter_range("imf")
        <DateTimeRange: 1997-09-02T00:00:12+00:00 -> ...>
        """
        return self._parameter_range(parameter_id)

    def dataset_range(self, dataset_id: str or DatasetIndex) -> Optional[DateTimeRange]:
        """Get dataset time range as defined in the inventory.

        Parameters
        ----------
        dataset_id: str or DatasetIndex
            parameter id

        Returns
        -------
        Optional[DateTimeRange]
            Data time range

        Examples
        --------
        >>> import speasy as spz
        >>> spz.amda.dataset_range("ace-imf-all")
        <DateTimeRange: 1997-09-02T00:00:12+00:00 -> ...>
        """
        return self._dataset_range(dataset_id)

    def is_user_catalog(self, catalog_id: str or CatalogIndex):
        """Tels if a product is a user catalog

        Parameters
        ----------
        catalog_id: str or CatalogIndex
            product id

        Returns
        -------
        bool
            True if the product is a user catalog, False otherwise.

        """
        return ImpexProvider.is_user_product(catalog_id, flat_inventories.__dict__[self.provider_name].catalogs)

    def is_user_timetable(self, timetable_id: str or TimetableIndex):
        """Tels if a product is a user timetable

        Parameters
        ----------
        timetable_id: str or TimetableIndex
            product id

        Returns
        -------
        bool
            True if the product is a user timetable, False otherwise.

        """
        return ImpexProvider.is_user_product(timetable_id, flat_inventories.__dict__[self.provider_name].timetables)

    def is_user_parameter(self, parameter_id: str or ParameterIndex):
        """Tells if a product is a user parameter

        Parameters
        ----------
        parameter_id: str or ParameterIndex
            product id

        Returns
        -------
        bool
            True if the product is a user parameter, False otherwise.

        """
        return ImpexProvider.is_user_product(parameter_id, flat_inventories.__dict__[self.provider_name].parameters)

    def get_data(self, product, start_time=None, stop_time=None,
                 **kwargs) -> MaybeAnyProduct:
        """Get product data by id, start and stop time if applicable. The product can be a parameter, a dataset, a
        catalog or a timetable. The method will automatically determine the product type and call the appropriate underlying
        method.

        Parameters
        ----------
        product: str or SpeasyIndex
            product id
        start_time: str or datetime.datetime
            desired data start time
        stop_time: str datetime.datetime
            desired data stop time

        Returns
        -------
        MaybeAnyProduct
            product data if available, None otherwise

        Examples
        --------

        >>> import speasy as spz
        >>> imf_data = spz.amda.get_data("imf", "2019-02-24T19:20:05", "2019-02-25")
        >>> print(imf_data.columns)
        ['bx', 'by', 'bz']
        >>> print(imf_data.values.shape)
        (1050, 3)


        """
        product_t = self.product_type(product)
        if product_t == ImpexProductType.DATASET and start_time and stop_time:
            return self.get_dataset(dataset_id=product, start=start_time, stop=stop_time, **kwargs)
        if product_t == ImpexProductType.PARAMETER and start_time and stop_time:
            if self.is_user_parameter(product):
                return self.get_user_parameter(parameter_id=product,
                                               start_time=start_time, stop_time=stop_time, **kwargs)
            else:
                return self.get_parameter(product=product, start_time=start_time, stop_time=stop_time, **kwargs)
        if product_t == ImpexProductType.CATALOG:
            if self.is_user_catalog(product):
                return self.get_user_catalog(catalog_id=product, **kwargs)
            else:
                return self.get_catalog(catalog_id=product, **kwargs)
        if product_t == ImpexProductType.TIMETABLE:
            if self.is_user_timetable(product):
                return self.get_user_timetable(timetable_id=product, **kwargs)
            else:
                return self.get_timetable(timetable_id=product, **kwargs)
        raise ValueError(f"Unknown product: {product}")

    def get_user_parameter(self, parameter_id: str or ParameterIndex, start_time: datetime or str,
                           stop_time: datetime or str, **kwargs) -> Optional[SpeasyVariable]:
        """Get user parameter. Raises an exception if user is not authenticated.

        Parameters
        ----------
        parameter_id: str or ParameterIndex
            parameter id
        start_time: datetime or str
            begining of data time
        stop_time: datetime or str
            end of data time

        Returns
        -------
        Optional[SpeasyVariable]
            user parameter

        Examples
        --------

        >>> import speasy as spz
        >>> user_param = spz.amda.get_user_parameter("ws_0", "2019-02-24T19:20:05", "2019-02-25") # doctest: +SKIP
        >>> print(user_param.columns) # doctest: +SKIP
        ['ws_test_param']
        >>> print(user_param.values.shape) # doctest: +SKIP
        (2, 1)


        Warnings
        --------
            Calling :meth:`~speasy.amda.amda.AMDA_Webservice.get_user_parameter` without having defined AMDA_Webservice
            login credentials will result in a :class:`~speasy.config.exceptions.UndefinedConfigEntry`
            exception being raised.

        """
        parameter_id = to_xmlid(parameter_id)
        start_time, stop_time = make_utc_datetime(start_time), make_utc_datetime(stop_time)
        return self._dl_user_parameter(parameter_id=parameter_id, start_time=start_time, stop_time=stop_time, **kwargs)

    def get_parameter(self, product, start_time, stop_time,
                      extra_http_headers: Dict or None = None,
                      output_format: str or None = None, **kwargs) -> Optional[SpeasyVariable]:
        """Get parameter data.

        Parameters
        ----------
        product: str or ParameterIndex
            parameter id
        start_time:
            desired data start time
        stop_time:
            desired data stop time
        extra_http_headers: dict
            reserved for internal use
        output_format: str
            request output format in case of success, allowed values are ASCII and CDF_ISTP

        Returns
        -------
        Optional[SpeasyVariable]
            product data if available

        Examples
        --------

        >>> import speasy as spz
        >>> import datetime
        >>> imf_data = spz.amda.get_parameter("imf", "2018-01-01", "2018-01-01T01")
        >>> print(imf_data.columns)
        ['bx', 'by', 'bz']
        >>> print(imf_data.values.shape)
        (225, 3)

        """
        if hasattr(self, 'has_time_restriction') and self.has_time_restriction(product, start_time, stop_time):
            kwargs['disable_proxy'] = True
            kwargs['restricted_period'] = True

        return self._get_parameter(product, start_time, stop_time, extra_http_headers=extra_http_headers,
                                   output_format=output_format or self.client.output_format, **kwargs)

    def get_dataset(self, dataset_id: str or DatasetIndex, start: str or datetime, stop: str or datetime,
                    **kwargs) -> Dataset or None:
        """Get dataset contents. Returns list of SpeasyVariable objects, one for each
        parameter in the dataset.

        Parameters
        ----------
        dataset_id: str or DatasetIndex
            dataset id
        start: str or datetime
            desired data start
        stop: str or datetime
            desired data end

        Returns
        -------
        Dataset or None
            dataset content as a collection of SpeasyVariable if it succeeds or None

        Examples
        --------

        >>> import speasy as spz
        >>> import datetime
        >>> dataset = spz.amda.get_dataset("ace-imf-all", datetime.datetime(2000,1,1), datetime.datetime(2000,1,2))
        >>> dataset
        <Dataset: final / prelim
                variables: ['|b|', 'b_gse', 'b_gsm']
                time range: <DateTimeRange: 2000-01-01T00:00:11+00:00 -> 2000-01-01T23:59:55+00:00>


        """
        ds_range = self.dataset_range(dataset_id)
        if not ds_range.intersect(DateTimeRange(start, stop)):
            log.warning(f"You are requesting {dataset_id} outside of its definition range {ds_range}")
            return None

        dataset_id = to_xmlid(dataset_id)
        name = self.flat_inventory.datasets[dataset_id].name
        meta = {k: v for k, v in self.flat_inventory.datasets[dataset_id].__dict__.items() if
                not isinstance(v, SpeasyIndex)}
        parameters = self.list_parameters(dataset_id)
        return Dataset(name=name,
                       variables={p.name: self.get_parameter(p, start, stop, **kwargs) for p in parameters},
                       meta=meta)

    def get_timetable(self, timetable_id: str or TimetableIndex, **kwargs) -> Optional[TimeTable]:
        """Get timetable data by ID.

        Parameters
        ----------
        timetable_id: str or TimetableIndex
            time table id

        Returns
        -------
        Optional[TimeTable]
            timetable data

        Examples
        --------

        >>> import speasy as spz
        >>> spz.amda.get_timetable("sharedtimeTable_0")
        <TimeTable: FTE_c1>

        """
        return self._dl_timetable(to_xmlid(timetable_id), **kwargs)

    def get_catalog(self, catalog_id: str or CatalogIndex, **kwargs) -> Optional[Catalog]:
        """Get catalog data by ID.

        Parameters
        ----------
        catalog_id: str or CatalogIndex
            catalog id

        Returns
        -------
        Optional[Catalog]
            catalog data

        Examples
        --------

        >>> import speasy as spz
        >>> spz.amda.get_catalog("sharedcatalog_22")
        <Catalog: model_regions_plasmas_mms_2019>

        """
        return self._dl_catalog(to_xmlid(catalog_id), **kwargs)

    def get_user_timetable(self, timetable_id: str or TimetableIndex, **kwargs) -> Optional[TimeTable]:
        """Get user timetable. Raises an exception if user is not authenticated.

        Parameters
        ----------
        timetable_id: str
            timetable id

        Returns
        -------
        Optional[TimeTable]
            user timetable

        Examples
        --------
        >>> import speasy as spz
        >>> spz.amda.get_user_timetable("tt_0") # doctest: +SKIP
        <TimeTable: test_alexis>

        Warnings
        --------
            Calling :meth:`~speasy.amda.amda.AMDA_Webservice.get_user_timetable` without having defined AMDA_Webservice
            login credentials will result in a :class:`~speasy.config.exceptions.UndefinedConfigEntry`
            exception being raised.

        """
        timetable_id = to_xmlid(timetable_id)
        return self._dl_user_timetable(to_xmlid(timetable_id), **kwargs)

    def get_user_catalog(self, catalog_id: str or CatalogIndex, **kwargs) -> Optional[Catalog]:
        """Get user catalog. Raises an exception if user is not authenticated.


        Parameters
        ----------
        catalog_id: str or CatalogIndex
            catalog id

        Returns
        -------
        Optional[Catalog]
            user catalog

        Examples
        --------

        >>> import speasy as spz
        >>> spz.amda.get_user_catalog("tt_0") # doctest: +SKIP
        <Catalog: test_alexis>

        Warnings
        --------
            Calling :meth:`~speasy.amda.amda.AMDA_Webservice.get_user_catalog` without having defined AMDA_Webservice
            login credentials will result in a :class:`~speasy.config.exceptions.UndefinedConfigEntry`
            exception being raised.

        """
        catalog_id = to_xmlid(catalog_id)
        return self._dl_user_catalog(catalog_id, **kwargs)

    def product_type(self, product_id: str or SpeasyIndex) -> ImpexProductType:
        """Returns product type for any known Impex product from its index or ID.

        Parameters
        ----------
        product_id: str or SpeasyIndex
            product id

        Returns
        -------
        ImpexProductType
            Type of product IE ImpexProductType.DATASET, ImpexProductType.TIMETABLE, ...

        Examples
        --------

        >>> import speasy as spz
        >>> spz.amda.product_type("imf")
        <ImpexProductType.PARAMETER: 2>
        >>> spz.amda.product_type("ace-imf-all")
        <ImpexProductType.DATASET: 1>
        """
        product_id = to_xmlid(product_id)
        if product_id in flat_inventories.__dict__[self.provider_name].datasets:
            return ImpexProductType.DATASET
        if product_id in flat_inventories.__dict__[self.provider_name].parameters:
            return ImpexProductType.PARAMETER
        if product_id in flat_inventories.__dict__[self.provider_name].components:
            return ImpexProductType.COMPONENT
        if product_id in flat_inventories.__dict__[self.provider_name].timetables:
            return ImpexProductType.TIMETABLE
        if product_id in flat_inventories.__dict__[self.provider_name].catalogs:
            return ImpexProductType.CATALOG

        return ImpexProductType.UNKNOWN

    def to_index(self, product_id: str or SpeasyIndex) -> AnyProductIndex:
        if type(product_id) in (
            DatasetIndex, ParameterIndex, TemplatedParameterIndex, ComponentIndex, TimetableIndex, CatalogIndex):
            return product_id
        elif type(product_id) is str:
            if p := flat_inventories.__dict__[self.provider_name].datasets.get(product_id):
                return p
            if p := flat_inventories.__dict__[self.provider_name].parameters.get(product_id):
                return p
            if p := flat_inventories.__dict__[self.provider_name].components.get(product_id):
                return p
            if p := flat_inventories.__dict__[self.provider_name].timetables.get(product_id):
                return p
            if p := flat_inventories.__dict__[self.provider_name].catalogs.get(product_id):
                return p
        raise ValueError(f"Unknown product: {product_id}")

    def find_parent_dataset(
        self,
        product_id: Union[str, DatasetIndex, ParameterIndex, TemplatedParameterIndex, ComponentIndex]
    ) -> Optional[str]:

        product_id = to_xmlid(product_id)
        product = self.to_index(product_id)
        if isinstance(product, DatasetIndex):
            return product_id
        elif type(product) in (ParameterIndex, ComponentIndex, TemplatedParameterIndex):
            return product.dataset
        return None

    @staticmethod
    def is_user_product(product_id: str or SpeasyIndex, collection: Dict):
        xmlid = to_xmlid(product_id)
        if xmlid in collection:
            return is_private(collection[xmlid])
        return False

    def list_datasets(self) -> List[DatasetIndex]:
        """Get the list of datasets available

        Returns
        -------
        List[DatasetIndex]
            list of dataset indexes

        Examples
        --------

        >>> import speasy as spz
        >>> amda_datasets = spz.amda.list_datasets()
        >>> len(amda_datasets) > 0
        True
        >>> amda_datasets[0]
        <DatasetIndex: ...>
        >>> amda_datasets[0].desc
        '...'

        """
        return list(filter(is_public, flat_inventories.__dict__[self.provider_name].datasets.values()))

    def list_parameters(self, dataset_id: Optional[str or DatasetIndex] = None) -> List[ParameterIndex]:
        """Get the list of parameter indexes available in AMDA or a given dataset

        Parameters
        ----------
        dataset_id: Optional[str or AMDADatasetIndex]
            optional parent dataset id

        Returns
        -------
        List[AMDAParameterIndex]
            the list of parameter indexes

        Examples
        --------

        >>> import speasy as spz
        >>> amda_parameters = spz.amda.list_parameters()
        >>> len(amda_parameters) > 0
        True
        >>> amda_parameters[0]
        <ParameterIndex: ...>


        """
        if dataset_id is not None:
            return list(flat_inventories.__dict__[self.provider_name].datasets[to_xmlid(dataset_id)])
        return list(filter(is_public, flat_inventories.__dict__[self.provider_name].parameters.values()))

    def list_user_parameters(self) -> List[ParameterIndex]:
        """Get the list of user parameters. User parameters are represented as dictionary objects.

        Returns
        -------
        List[AMDAParameterIndex]
            list of user parameters

        Examples
        --------

        >>> import speasy as spz
        >>> spz.amda.list_user_parameters() # doctest: +SKIP
        [<ParameterIndex: test_param>]

        Warnings
        --------
           Calling :meth:`~speasy.amda.amda.AMDA_Webservice.get_user_parameter` without having defined AMDA_Webservice
           login credentials will result in a :class:`~speasy.config.exceptions.UndefinedConfigEntry`
           exception being raised.


        """
        return list(filter(is_private, flat_inventories.__dict__[self.provider_name].parameters.values()))

    def list_timetables(self) -> List[TimetableIndex]:
        """Get list of public timetables.

        Returns
        -------
        List[AMDATimetableIndex]
            list of timetable IDs.

        Examples
        --------

        >>> import speasy as spz
        >>> spz.amda.list_timetables()[::50]
        [<TimetableIndex: ...>, <TimetableIndex: ...>, <TimetableIndex: ...>]

        """
        return list(filter(is_public, flat_inventories.__dict__[self.provider_name].timetables.values()))

    def list_user_timetables(self) -> List[TimetableIndex]:
        """Get the list of user timetables. User timetable are represented as dictionary objects.

        Returns
        -------
        List[AMDATimetableIndex]
            list of user timetables.

        Examples
        --------

        >>> import speasy as spz
        >>> spz.amda.list_user_timetables() # doctest: +SKIP
        [<TimetableIndex: test_alexis>, <TimetableIndex: test_alexis2>, <TimetableIndex: tt3>]

        Warnings
        --------
           Calling :meth:`~speasy.amda.amda.AMDA_Webservice.get_user_timetables` without having defined AMDA_Webservice
           login credentials will result in a :class:`~speasy.config.exceptions.UndefinedConfigEntry`
           exception being raised.


        """
        return list(filter(is_private, flat_inventories.__dict__[self.provider_name].timetables.values()))

    def list_catalogs(self) -> List[CatalogIndex]:
        """Get the list of public catalog IDs:

        Returns
        -------
        List[AMDACatalogIndex]
            list of catalog IDs

        Examples
        --------

        >>> import speasy as spz
        >>> amda_catalogs = spz.amda.list_catalogs()
        >>> len(amda_catalogs) > 0
        True
        >>> amda_catalogs[0]
        <CatalogIndex: model_regions_plasmas_mms_2019>

        """
        return list(filter(is_public, flat_inventories.__dict__[self.provider_name].catalogs.values()))

    def list_user_catalogs(self) -> List[CatalogIndex]:
        """Get the list of user catalogs. User catalogs are represented as dictionary objects.

        Returns
        -------
        List[AMDACatalogIndex]
            list of user catalogs.

        Examples
        --------

        >>> import speasy as spz
        >>> spz.amda.list_user_catalogs() # doctest: +SKIP
        [<CatalogIndex: MyCatalog>]

        Warnings
        --------
           Calling :meth:`~speasy.amda.amda.AMDA_Webservice.get_user_catalogs` without having defined AMDA_Webservice
           login credentials will result in a :class:`~speasy.config.exceptions.UndefinedConfigEntry`
           exception being raised.


        """
        return list(filter(is_private, flat_inventories.__dict__[self.provider_name].catalogs.values()))

    def _get_parameter(self, product, start_time, stop_time,
                       extra_http_headers: Dict or None = None, output_format: str or None = None,
                       restricted_period=False, **kwargs) -> \
        Optional[
            SpeasyVariable]:
        log.debug(f'Get data: product = {product}, data start time = {start_time}, data stop time = {stop_time}')
        if hasattr(self, 'get_real_product_id'):
            real_product_id = self.get_real_product_id(product, **kwargs)
            if real_product_id:
                kwargs['real_product_id'] = real_product_id
        return self._dl_parameter(start_time=start_time, stop_time=stop_time, parameter_id=product,
                                  extra_http_headers=extra_http_headers,
                                  output_format=output_format,
                                  product_variables=self._get_product_variables(product, **kwargs),
                                  restricted_period=restricted_period,
                                  time_format='UNIXTIME', **kwargs)

    def _dl_parameter_chunk(self, start_time: datetime, stop_time: datetime, parameter_id: str,
                            extra_http_headers: Dict or None = None,
                            use_credentials: bool = False,
                            product_variables: List = None, **kwargs) -> Optional[SpeasyVariable]:
        url = self.client.get_parameter(start_time=start_time.strftime('%Y-%m-%dT%H:%M:%SZ'),
                                        stop_time=stop_time.strftime('%Y-%m-%dT%H:%M:%SZ'),
                                        parameter_id=parameter_id,
                                        extra_http_headers=extra_http_headers,
                                        use_credentials=use_credentials, **kwargs)
        # check status until done
        if url is not None:
            var = None
            if not product_variables:
                product_variables = [kwargs.get('real_product_id', parameter_id)]
            if kwargs.get('output_format', self.client.output_format) in ["CDF_ISTP", "CDF"]:
                var = self._cdf_codec.load_variables(variables=product_variables, file=url)
            else:
                raise NotImplementedError(f"Output format {kwargs.get('output_format')} not supported")
            if var is not None:
                if isinstance(var, SpeasyVariable):
                    if len(var):
                        log.debug(
                            f'Loaded var: data shape = {var.values.shape}, data start time = {var.time[0]}, \
                                    data stop time = {var.time[-1]}')
                    else:
                        log.debug('Loaded var: Empty var')
                else:
                    if parameter_id in self.flat_inventory.parameters:
                        name = self.flat_inventory.parameters[parameter_id].spz_name()
                    else:
                        name = parameter_id
                    var = ImpexProvider._concatenate_variables(var, name)
                    if var is None:
                        log.debug('Failed to concatenate variables')
            else:
                log.debug(f'Failed to load file f{url}')
            return var
        return None

    def _dl_parameter(self, start_time: datetime, stop_time: datetime, parameter_id: str,
                      extra_http_headers: Dict or None = None, restricted_period=False,
                      use_credentials: bool = False,
                      product_variables: List = None, **kwargs) -> Optional[SpeasyVariable]:
        dt = timedelta(days=self.max_chunk_size_days)
        if restricted_period:
            if not self.client.credential_are_valid():
                raise MissingCredentials(
                    "Restricted period requested but no credentials provided, please add your "
                    "{} credentials.".format(self.provider_name))
            else:
                use_credentials = True
        if stop_time - start_time > dt:
            var = None
            curr_t = start_time
            while curr_t < stop_time:
                var = merge([var, self._dl_parameter_chunk(curr_t, min(curr_t + dt, stop_time), parameter_id,
                                                           extra_http_headers=extra_http_headers,
                                                           product_variables=product_variables, **kwargs)])
                curr_t += dt
            return var
        else:
            return self._dl_parameter_chunk(start_time, stop_time, parameter_id,
                                            extra_http_headers=extra_http_headers,
                                            use_credentials=use_credentials,
                                            product_variables=product_variables, **kwargs)

    def _dl_user_parameter(self, start_time: datetime, stop_time: datetime, parameter_id: str,
                           **kwargs) -> Optional[SpeasyVariable]:
        return self._dl_parameter(parameter_id=parameter_id, start_time=start_time, stop_time=stop_time,
                                  product_variables=self._get_product_variables(parameter_id, **kwargs),
                                  use_credentials=True, **kwargs)

    def _dl_timetable(self, timetable_id: str, use_credentials=False, **kwargs):
        get_timetable_url = self.client.get_timetable(timetable_id, use_credentials=use_credentials, **kwargs)
        if get_timetable_url is not None:
            timetable = load_timetable(filename=get_timetable_url)
            if timetable:
                timetable.meta.update(
                    flat_inventories.__dict__[self.provider_name].timetables.get(timetable_id,
                                                                                 SimpleNamespace()).__dict__)
                log.debug(f'Loaded timetable: id = {timetable_id}')  # lgtm[py/clear-text-logging-sensitive-data]
            else:
                log.debug('Got None')
            return timetable
        return None

    def _dl_user_timetable(self, timetable_id: str, **kwargs):
        return self._dl_timetable(timetable_id, use_credentials=True, **kwargs)

    def _dl_catalog(self, catalog_id: str, use_credentials=False, **kwargs):
        get_catalog_url = self.client.get_catalog(catalog_id, use_credentials=use_credentials, **kwargs)
        if get_catalog_url is not None:
            catalog = load_catalog(get_catalog_url)
            if catalog:
                log.debug(f'Loaded catalog: id = {catalog_id}')  # lgtm[py/clear-text-logging-sensitive-data]
                catalog.meta.update(
                    flat_inventories.__dict__[self.provider_name].catalogs.get(catalog_id, SimpleNamespace()).__dict__)
            else:
                log.debug('Got None')
            return catalog
        return None

    def _dl_user_catalog(self, catalog_id: str, **kwargs):
        return self._dl_catalog(catalog_id, use_credentials=True, **kwargs)

    def _get_product_variables(self, product_id: str or SpeasyIndex, **kwargs):
        product_id = to_xmlid(product_id)
        return [kwargs.get('real_product_id', product_id)]

    @staticmethod
    def _concatenate_variables(variables: Dict[str, SpeasyVariable], product_id) -> Optional[SpeasyVariable]:
        if len(variables) == 0:
            return None
        elif len(variables) == 1:
            return list(variables.values())[0]

        axes = []
        columns = []
        values = None
        meta = {}
        for name, variable in variables.items():
            if not axes:
                values = variable.values.copy()
                axes = variable.axes.copy()
                meta = deepcopy(variable.meta)
            else:
                values = np.concatenate((values, variable.values), axis=1)
                axes += [axis.copy() for axis in variable.axes[1:]]
            columns.append(name)

        if 'FIELDNAM' in meta:
            meta['FIELDNAM'] = product_id

        return SpeasyVariable(
            axes=axes,
            values=DataContainer(values=values, meta=meta, name=product_id, is_time_dependent=True),
            columns=columns)

    def _get_obs_data_tree(self, add_template_info=False) -> str or None:
        return self.client.get_obs_data_tree(add_template_info=add_template_info)

    def _get_timetables_tree(self) -> str or None:
        return self.client.get_time_table_list()

    def _get_catalogs_tree(self) -> str or None:
        return self.client.get_catalog_list()

    def _get_user_timetables_tree(self) -> str or None:
        return self.client.get_time_table_list(use_credentials=True)

    def _get_user_catalogs_tree(self) -> str or None:
        return self.client.get_catalog_list(use_credentials=True)

    def _get_derived_parameter_tree(self) -> str or None:
        return self.client.get_derived_parameter_list()



from typing import Union
import datetime
import dateutil
from packaging.version import Version, parse, InvalidVersion


def _str_to_version_datetime(v: str) -> datetime.datetime or None:
    try:
        version = dateutil.parser.parse(v)
    except ValueError:
        version = None
    return version


def str_to_version(v: str) -> Version or datetime.datetime or None:
    """Converts given version str representation to a compatible version object

    Parameters
    ----------
    v: str
        version value to convert

    Returns
    -------
    Version or datetime.datetime or None
        a compatible version object or None

    Examples
    --------

    >>> from speasy.core.cache.version import str_to_version
    >>> import datetime
    >>> str_to_version("1.2.3")
    <Version('1.2.3')>
    >>> str_to_version('2010-01-01T00:00:00')
    datetime.datetime(2010, 1, 1, 0, 0)

    See Also
    --------
    version_to_str
    """
    try:
        version = parse(v)
        if "LegacyVersion" in str(type(version)):
            return _str_to_version_datetime(v)
    except InvalidVersion:
        return _str_to_version_datetime(v)
    return version


def version_to_str(v: Union[Version, datetime.datetime]) -> str:
    """Converts given version value to its str representation

    Parameters
    ----------
    v: Version or datetime.datetime
        version value to convert

    Returns
    -------
    str
        string representation of given version

    Examples
    --------

    >>> from speasy.core.cache.version import version_to_str
    >>> from packaging.version import Version
    >>> import datetime
    >>> version_to_str(Version("1.2.3"))
    '1.2.3'
    >>> version_to_str(datetime.datetime(2010, 1, 1))
    '2010-01-01T00:00:00'

    See Also
    --------
    str_to_version
    """
    if type(v) is Version:
        return str(v)
    elif type(v) is datetime.datetime:
        return v.isoformat()



"""
speasy.config
-------------

Configuration module for SPEASY, it reads or sets config entries first from ENV then from config file.
"""

import configparser
import os
from typing import Any
import ast

import appdirs

from ..core import mkdir

SPEASY_CONFIG_DIR = str(appdirs.user_config_dir(appname="speasy", appauthor="LPP"))

SPEASY_CONFIG_FILE = os.path.join(SPEASY_CONFIG_DIR, "config.ini")
mkdir(os.path.dirname(SPEASY_CONFIG_FILE))
_config = configparser.ConfigParser()
_config.read(SPEASY_CONFIG_FILE)

_entries = {}


def _load_dict_from_repr(value: str):
    if value:
        d = ast.literal_eval(value)
        if isinstance(d, dict):
            return d
        else:
            raise ValueError(f"Config value can't be converted to dict: {value}")
    return {}


def show():
    """Prints config entries and current values
    """
    for section in _entries.values():
        print(section)


def _save_changes():
    with open(SPEASY_CONFIG_FILE, 'w') as f:
        _config.write(f)


class ConfigEntry:
    """Configuration entry class. Used to set and get configuration values.

    Attributes
    ----------
    key1: str
        Module or category name
    key2: str
        Entry name
    default: str
        Default value
    type_ctor: Any
        function called to get value from string repr
    env_var_name: str
        Environment variable name to use to set this entry

    Methods
    -------
    get:
        Get entry current value
    set:
        Set entry value (could be env or file)
    """

    def __init__(self, key1: str, key2: str, default: Any = "", type_ctor=None, description: str = ""):
        self.key1 = key1
        self.key2 = key2
        self.default = str(default)
        self.type_ctor = type_ctor or (lambda x: x)
        self.description = description
        self.env_var_name = f"SPEASY_{self.key1}_{self.key2}".upper().replace(
            '-', '_')

    def __repr__(self):
        return f"""ConfigEntry: {self.key1}/{self.key2}
    environment variable name: {self.env_var_name}
    value:                     {self.get()}
    description:               {self.description}"""

    def get(self):
        """Get configuration entry value. If a default is not provided then raise :class:`~speasy.config.exceptions.UndefinedConfigEntry`.

        Returns
        -------
        str:
            configuration value
        """
        if self.env_var_name in os.environ:
            return self.type_ctor(os.environ[self.env_var_name])
        if self.key1 in _config and self.key2 in _config[self.key1]:
            return self.type_ctor(_config[self.key1][self.key2])
        return self.type_ctor(self.default)

    def set(self, value: str):
        if self.env_var_name in os.environ:
            os.environ[self.env_var_name] = str(value)
        if self.key1 not in _config:
            _config.add_section(self.key1)
        _config[self.key1][self.key2] = str(value)
        _save_changes()

    def __call__(self, *args, **kwargs):
        return self.get()


class ConfigSection:
    def __init__(self, name, **kwargs):
        self.__dict__.update({
            entry_name: ConfigEntry(name, entry_name, **e_kwargs) for entry_name, e_kwargs in kwargs.items()
        })
        _entries[name] = self
        self.name = name

    def __repr__(self):
        s = f"""
============================================
\t\t{self.name}
============================================"""
        for _, entry in self.__dict__.items():
            s += f"\n  {entry}\n-------------------------------------------"
        return s


def remove_entry(entry: ConfigEntry):
    """Deletes entry from config file and its section if it was the last entry

    Parameters
    ----------
    entry: ConfigEntry
        the entry to delete

    Returns
    -------
    None

    """
    if entry.key1 in _config:
        section = _config[entry.key1]
        if entry.key2 in section:
            section.pop(entry.key2)
        if len(section) == 0:
            _config.remove_section(entry.key1)
        _save_changes()


# ==========================================================================================
#                           ADD HERE CONFIG ENTRIES
# user can easily discover them with speasy.config.<completion>
# ==========================================================================================
core = ConfigSection("CORE",
                     disabled_providers={"default": set(),
                                         "description": """A comma separated list of providers you want to disable.
The main benefit of disabling providers is to speedup speasy loading.""",
                                         "type_ctor": lambda x: set(x.split(','))},
                     http_rewrite_rules={"default": {
                         "https://cdaweb.gsfc.nasa.gov/pub/": "http://sciqlop.lpp.polytechnique.fr/cdaweb-data/pub/"},
                         "description": """A dictionary of rules to rewrite URLs before sending requests.
The keys are the URL to match and the values are the replacement URL.
Example: {"http://example.com": "http://localhost:8000"}""",
                         "type_ctor": _load_dict_from_repr},
                     urlib_pool_size={"default": 10,
                                      "description": """Sets the maximum number of connections to keep in the pool.
This is useful to avoid creating a new connection for each request.""",
                                      "type_ctor": int},
                     urlib_num_pools={"default": 10,
                                      "description": """Sets the maximum number of pools to keep in the pool.
This is useful to avoid creating a new pool for each request.""",
                                      "type_ctor": int},
                     user_codecs_extra_dirs={"default": set(),
                                             "description": """A comma separated list of directories to scan for extra codecs.""",
                                             "type_ctor": lambda x: set(x.split(','))},
                     )

proxy = ConfigSection("PROXY",
                      enabled={"default": True,
                               "description": """Enables or disables speasy proxy usage.
Speasy proxy is an intermediary server which helps by caching requests among several users.""",
                               "type_ctor": lambda x: {'true': True, 'false': False}.get(x.lower(), False)},
                      url={"default": "http://sciqlop.lpp.polytechnique.fr/cache",
                           "description": """Speasy proxy server URL, you can use http://sciqlop.lpp.polytechnique.fr/cache.
Speasy proxy is an intermediary server which helps by caching requests among several users."""}
                      )

cache = ConfigSection("CACHE",
                      size={"default": 20e9, "description": """Sets the maximum cache capacity.""",
                            "type_ctor": lambda x: int(float(x))},
                      path={"default": str(appdirs.user_cache_dir("speasy", "LPP")),
                            "description": """Sets Speasy cache path."""}
                      )

index = ConfigSection("INDEX",
                      path={
                          "default": f'{appdirs.user_data_dir("speasy", "LPP")}/index'}
                      )
cdaweb = ConfigSection("CDAWEB",
                       inventory_data_path={
                           "default": f'{appdirs.user_data_dir("speasy", "LPP")}/cda_inventory'},
                       preferred_access_method={"default": "BEST",
                                                "description": "Preferred access method to CDAWeb, either REST API (API), dirtect files download (FILE) or BEST to let Speasy choose the fastest and most likely to work method based on the requested product."},
                       )

amda = ConfigSection("AMDA",
                     username={
                         "description": """Your AMDA username, once set, you will be able to get your private products."""},
                     password={
                         "description": """Your AMDA password, once set, you will be able to get your private products."""},
                     user_cache_retention={"default": 900,
                                           "description": "AMDA specific cache retention for requests such as list_catalogs.",
                                           "type_ctor": int
                                           },
                     max_chunk_size_days={
                         "default": 10,
                         "description": "Maximum request duration in days, any request over a longer period will be split into smaller ones.",
                         "type_ctor": int},
                     entry_point={
                         "default": "https://amda.irap.omp.eu"},
                     output_format={
                         "description": "File format requested to AMDA, only CDF_ISTP is supported for now.",
                         "default": "CDF_ISTP"}
                     )

archive = ConfigSection("ARCHIVE",
                        extra_inventory_lookup_dirs={"default": set(),
                                                     "description": """A comma separated list of directory path Archive provider will scann for YAML inventory files.""",
                                                     "type_ctor": lambda x: set(x.split(','))}
                        )

inventories = ConfigSection("INVENTORIES",
                            cache_retention_days={
                                "default": 2,
                                "description": "Maximum times in days speasy will keep inventories in cache before fetching newer version.",
                                "type_ctor": int}
                            )



"""
speasy.amda.amda
----------------

This package contains the definition of the :class:`~speasy.data_providers.amda.ws.AMDA_Webservice` class, the object that
manages the connexion to `AMDA <http://amda.irap.omp.eu/>`_ and allows users to list available products, get their
description and downloading the corresponding data.

AMDA provides the following kinds products :
    - parameters, measurements as timeseries
        - list them with :meth:`~speasy.data_providers.amda.ws.AMDA_Webservice.get_parameter()`
        - download them with :meth:`~speasy.data_providers.amda.ws.AMDA_Webservice.list_parameters()`
    - datasets, collections of parameters with the same time axis
        - list them with :meth:`~speasy.data_providers.amda.ws.AMDA_Webservice.get_dataset()`
        - download them with :meth:`~speasy.data_providers.amda.ws.AMDA_Webservice.list_datasets()`
    - timetables, lists of time ranges
        - list them with :meth:`~speasy.data_providers.amda.ws.AMDA_Webservice.get_timetable()`
        - download them with :meth:`~speasy.data_providers.amda.ws.AMDA_Webservice.list_timetables()`
    - catalogs, timetables with metadata for each time range
        - list them with :meth:`~speasy.data_providers.amda.ws.AMDA_Webservice.get_catalog()`
        - download them with :meth:`~speasy.data_providers.amda.ws.AMDA_Webservice.list_catalogs()`

See :doc:`user documentation </user/amda/amda>` for a more accessible documentation.

Notes
-----
You should not create an instance of  :class:`~speasy.data_providers.amda.ws.AMDA_Webservice` but use the `speasy.amda`
instance instead.

"""

__author__ = """Alexis Jeandet"""
__email__ = 'alexis.jeandet@member.fsf.org'
__version__ = '0.1.0'

from .ws import AmdaWebservice



from scipy import signal
from typing import Callable, Union, Collection
from speasy.products import SpeasyVariable
import numpy as np


def _apply_filter(filter_function: Callable, sos: np.ndarray, var: SpeasyVariable) -> SpeasyVariable:
    res = np.empty_like(var)
    res.values[:] = filter_function(sos=sos, x=var.values, axis=0)
    return res


def apply_sos_filter(sos: np.ndarray, filter_function: Callable,
                     var: Union[SpeasyVariable, Collection[SpeasyVariable]]) -> Union[
    SpeasyVariable, Collection[SpeasyVariable]]:
    """Apply an IIR filter to the variable(s) using the given filter function. This function just applies the filter to the
    values of the variable without any resampling, it assumes that the variable has a regular time axis.

    Parameters
    ----------
    sos: np.ndarray
        Second-order sections representation of the filter, as returned by :func:`scipy.signal.iirfilter` with `output='sos'` for example.
    filter_function: Callable
        The filter function to use (e.g. :func:`scipy.signal.sosfiltfilt`)
    var: SpeasyVariable or Collection[SpeasyVariable]
        The variable(s) to filter

    Returns
    -------
    SpeasyVariable or Collection[SpeasyVariable]
        The filtered variable(s)

    Notes
    -----
    It only supports 1D variables.
    """

    if isinstance(var, SpeasyVariable):
        return _apply_filter(filter_function, sos, var)
    else:
        return [_apply_filter(filter_function, sos, v) for v in var]


def sosfiltfilt(sos: np.ndarray, var: Union[SpeasyVariable, Collection[SpeasyVariable]]) -> Union[
    SpeasyVariable, Collection[SpeasyVariable]]:
    """Apply an IIR filter to the variable(s) using :func:`scipy.signal.sosfiltfilt`. This function just applies the filter to
    the values of the variable without any resampling, it assumes that the variable has a regular time axis.

    Parameters
    ----------
    sos: np.ndarray
        Second-order sections representation of the filter
    var: SpeasyVariable or Collection[SpeasyVariable]
        The variable(s) to filter

    Returns
    -------
    SpeasyVariable or Collection[SpeasyVariable]
        The filtered variable(s)

    Notes
    -----
    It only supports 1D variables.
    """
    return apply_sos_filter(sos, signal.sosfiltfilt, var)



from typing import Union, Optional
import io
from speasy.core.codecs.codec_interface import Buffer
from .csv_file import HapiCsvFile
import json
import pandas as pds


def _to_csv(hapi_csv_file: HapiCsvFile, dest:io.IOBase, with_headers=True) -> bool:
    if with_headers:
        headers = {
            "HAPI": "3.2",
            "status": {
                "code": 1200,
                "message": "OK request successful"
            },
            "parameters": [column.meta for column in hapi_csv_file.parameters]
        }
        dest.write(f'#{json.dumps(headers)}\n')
    data = {}
    for param in hapi_csv_file.parameters:
        data[param.name] = param.values
    df = pds.DataFrame(data)
    df.to_csv(dest, index=False, header=False, date_format='%Y-%m-%dT%H:%M:%S.%fZ', float_format='%.3g')
    return True





def save_hapi_csv(hapi_csv_file: HapiCsvFile, file: Optional[Union[str, io.IOBase]] = None) -> Union[bool, Buffer]:
    if type(file) is str:
        with open(file, 'wb') as f:
            return _to_csv(hapi_csv_file, f)
    elif hasattr(file, 'write'):
        return _to_csv(hapi_csv_file, file)
    elif file is None:
        buff = io.BytesIO()
        _to_csv(hapi_csv_file, buff)
        return buff.getvalue()
    raise ValueError("Invalid file type")





import logging
from datetime import timedelta
from typing import List, Optional, Union

import pyistp
from pyistp.loader import DataVariable, ISTPLoader
from speasy.core.any_files import any_loc_open
from speasy.core.cache import CacheCall
from speasy.core.inventory.indexes import ParameterIndex, DatasetIndex

log = logging.getLogger(__name__)


def filter_variable_meta(datavar: DataVariable) -> dict:
    keep_list = ['CATDESC', 'FIELDNAM', 'UNITS', 'UNIT_PTR', 'DISPLAY_TYPE', 'LABLAXIS', 'LABL_PTR_1', 'LABL_PTR_2',
                 'LABL_PTR_3', 'VIRTUAL', 'FUNCT', 'FILLVAL']
    base = {key: value for key, value in datavar.attributes.items() if key in keep_list}
    base['cdf_type'] = datavar.cdf_type
    if len(datavar.values.shape) == 1:
        base['spz_shape'] = 1
    else:
        base['spz_shape'] = datavar.values.shape[1:]
    return base


def _attribute_value(attr):
    if len(attr) == 1:
        return attr[0]
    else:
        return list(attr)


def filter_dataset_meta(dataset: ISTPLoader) -> dict:
    keep_list = ['Caveats', 'Rules_of_use', 'Time_resolution', 'spase_DatasetResourceID', 'HTTP_LINK', 'Data_type',
                 'Acknowledgement']
    return {key: _attribute_value(dataset.attribute(key)) for key in dataset.attributes() if key in keep_list}


def extract_parameter(cdf: ISTPLoader, var_name: str, provider: str, uid_fmt: str = "{var_name}", meta=None) -> \
    Optional[ParameterIndex]:
    try:
        datavar = cdf.data_variable(var_name)
        meta = meta or {}
        if datavar is not None:
            return ParameterIndex(name=var_name, provider=provider, uid=uid_fmt.format(var_name=var_name),
                                  meta={**filter_variable_meta(datavar), **meta})
    except IndexError or RuntimeError:
        print(f"Issue loading {var_name} from {cdf}")

    return None


def _extract_parameters_impl(cdf: ISTPLoader, provider: str, uid_fmt: str = "{var_name}", meta=None) -> List[
    ParameterIndex]:
    return list(filter(lambda p: p is not None,
                       map(lambda var_name: extract_parameter(cdf, var_name, provider, uid_fmt, meta=meta),
                           cdf.data_variables())))


def extract_parameters(url_or_istp_loader: Union[str,ISTPLoader], provider: str, uid_fmt: str = "{var_name}", meta=None) -> List[ParameterIndex]:
    indexes: List[ParameterIndex] = []
    try:
        if isinstance(url_or_istp_loader, str):
            with any_loc_open(url_or_istp_loader) as remote_cdf:
                cdf = pyistp.load(buffer=remote_cdf.read())
                return _extract_parameters_impl(cdf, provider=provider, uid_fmt=uid_fmt, meta=meta)
        else:
            return _extract_parameters_impl(url_or_istp_loader, provider=provider, uid_fmt=uid_fmt, meta=meta)

    except RuntimeError:
        print(f"Issue loading {url_or_istp_loader}")
    return indexes


@CacheCall(cache_retention=timedelta(days=7), is_pure=True)
def make_dataset_index(url: str, name: str, provider: str, uid: str, meta=None,
                       params_uid_format: str = "{var_name}", params_meta=None) -> Optional[DatasetIndex]:
    try:
        with any_loc_open(url, cache_remote_files=True) as remote_cdf:
            meta = meta or {}
            params_meta = params_meta or {}
            cdf = pyistp.load(buffer=remote_cdf.read())
            dataset = DatasetIndex(name=name, provider=provider, uid=uid, meta={**filter_dataset_meta(cdf), **meta})
            dataset.__dict__.update(
                {p.spz_name(): p for p in
                 _extract_parameters_impl(cdf, provider=provider, uid_fmt=params_uid_format, meta=params_meta)})
            return dataset
    except RuntimeError:
        print(f"Issue loading {url}")
    return None



import matplotlib.pyplot as plt
import matplotlib.colors as colors
import numpy as np


class Plot:

    def _get_ax(self, ax):
        # ax handling taken from pandas source code here:
        # https://github.com/pandas-dev/pandas/blob/main/pandas/plotting/_matplotlib/__init__.py#L64
        if ax is None and len(plt.get_fignums()) > 0:
            with plt.rc_context():
                ax = plt.gca()
            ax = getattr(ax, "left_ax", ax)
        if ax is None:
            fig, ax = plt.subplots()
        return ax

    def line(self, x, y, ax=None, labels=None, units=None, xaxis_label=None, yaxis_label=None, *args, **kwargs):
        ax = self._get_ax(ax)
        ax.tick_params(axis='x', labelrotation=45)
        ax.plot(x, y, label=labels)
        if labels is not None:
            ax.legend()
        if units is not None and yaxis_label is not None:
            ax.set_ylabel(f"{yaxis_label} ({units})")
        if xaxis_label is not None:
            ax.set_xlabel(f"{xaxis_label}")
        return ax

    def colormap(self, x, y, z, xaxis_label=None, yaxis_label=None, yaxis_units=None, zaxis_label=None,
                 zaxis_units=None, ax=None,
                 cmap=None, logy=True,
                 logz=True, vmin=None, vmax=None, *args,
                 **kwargs):
        ax = self._get_ax(ax)

        if yaxis_units is not None and yaxis_label is not None:
            ax.set_ylabel(f"{yaxis_label} ({yaxis_units})")
        if xaxis_label is not None:
            ax.set_xlabel(f"{xaxis_label}")

        vmin = vmin or np.nanmin(z[np.nonzero(z)])
        vmax = vmax or np.nanmax(z)

        if logy:
            ax.semilogy()
        if logz:
            norm = colors.LogNorm(vmin=vmin, vmax=vmax)
        else:
            norm = colors.Normalize(vmin=vmin, vmax=vmax)

        ax.tick_params(axis='x', labelrotation=45)
        cm = ax.pcolormesh(x, y, z,
                           cmap=cmap or 'plasma',
                           norm=norm, *args, **kwargs)
        cbar = plt.colorbar(cm, ax=ax)
        if zaxis_units is not None and zaxis_label is not None:
            cbar.set_label(f'{zaxis_label} ({zaxis_units})')
        return ax

    def __call__(self, *args, **kwargs):
        pass



from typing import List,  Optional, Dict, Any
import logging

import numpy as np
log = logging.getLogger(__name__)

class HapiCsvParameter:
    def __init__(self, values: np.ndarray, meta: Dict[str, Any]):
        self._meta = meta
        self._values = values

    @property
    def name(self):
        return self._meta["name"]

    @property
    def meta(self):
        return self._meta

    @property
    def values(self):
        return self._values


class HapiCsvFile:
    def __init__(self):
        self._parameters: List[HapiCsvParameter] = []

    def create_parameter(self, values: np.ndarray, meta: Dict[str, Any])->HapiCsvParameter:
        parameter = HapiCsvParameter(values, meta)
        self._parameters.append(parameter)
        return parameter

    def add_parameter(self, parameter: HapiCsvParameter):
        self._parameters.append(parameter)

    @property
    def time_axis(self):
        return self._parameters[0].values

    @property
    def time_axis_meta(self):
        return self._parameters[0].meta

    def get_parameter(self, name: str) -> Optional[HapiCsvParameter]:
        for par in self._parameters:
            if par.name == name:
                return par
        return None

    @property
    def parameters(self):
        return self._parameters



from typing import List, AnyStr, Optional, Mapping, Union, Dict, Any
import io
import logging

from datetime import timedelta
import numpy as np

from speasy.core.codecs import CodecInterface, register_codec
from speasy.core.codecs.codec_interface import Buffer
from speasy.core.cache import CacheCall
from speasy.products import SpeasyVariable, VariableTimeAxis, DataContainer
from .csv_file import HapiCsvFile, HapiCsvParameter

log = logging.getLogger(__name__)
from .reader import load_hapi_csv
from .writer import save_hapi_csv


def _create_meta(variable:SpeasyVariable) -> Dict[str, Any]:
    meta = {
        "name": variable.name,
        "units": variable.unit,
        "fill": variable.fill_value,
        "description": variable.meta.get("description", "")
    }
    if  np.issubdtype(variable.values.dtype, np.integer):
        meta["type"] = "int"
    elif np.issubdtype(variable.values.dtype, np.floating):
        meta["type"] = "double"
    else:
        raise ValueError(f"Unsupported data type {variable.values.dtype}")

    labels  =  variable.columns
    if labels is not None and len(labels) > 0:
        meta["label"] = labels

    if "coordinateSystemName" in variable.meta:
        meta["coordinateSystemName"] = variable.meta["coordinateSystemName"]
    if "vectorComponents" in variable.meta:
        meta["vectorComponents"] = variable.meta["vectorComponents"]

    if  len(variable.values.shape) > 1:
        meta["size"] = variable.values.shape[1:]
    return meta


def _decode_meta(meta: Dict[str, Any]) -> Dict[str, Any]:
    if "units" in meta:
        meta["UNITS"] = meta.pop("units")
    return meta


def _same_time_axis(variables: List[SpeasyVariable]) -> bool:
    if len(variables) < 2:
        return True
    ref_time_axis = variables[0].time
    return all([np.all(var.time == ref_time_axis) for var in variables[1:]])


def _hapi_csv_to_speasy_variables(hapi_csv_file: HapiCsvFile, variables: List[AnyStr]) -> Mapping[str, SpeasyVariable]:
    time_axis = VariableTimeAxis(values=hapi_csv_file.time_axis, meta=hapi_csv_file.time_axis_meta)
    loaded_vars = {}
    for var_name in variables:
        parameter = hapi_csv_file.get_parameter(var_name)
        if parameter is not None:
            loaded_vars[var_name] = SpeasyVariable(axes=[time_axis], values=DataContainer(parameter.values,
                                                                                          meta=_decode_meta(
                                                                                              parameter.meta)))
    return loaded_vars


def _make_hapi_csv_parameter(variable: SpeasyVariable) -> HapiCsvParameter:
    return HapiCsvParameter(values=variable.values.values,
                            meta=_create_meta(variable))


def _make_hapi_csv_time_axis(time_axis: VariableTimeAxis) -> HapiCsvParameter:
    return HapiCsvParameter(values=time_axis.values,
                            meta={"name": "Time", "type": "isotime", "units": "UTC", "length": 30, "fill": None})


def _speasy_variables_to_hapi_csv(variables: List[SpeasyVariable]) -> HapiCsvFile:
    if not _same_time_axis(variables):
        raise ValueError("All variables must have the same time axis")
    if len(variables) == 0:
        raise ValueError("No variables to save")
    hapi_csv_file = HapiCsvFile()
    hapi_csv_file.add_parameter(_make_hapi_csv_time_axis(variables[0].time))
    for var in variables:
        hapi_csv_file.add_parameter(_make_hapi_csv_parameter(var))
    return hapi_csv_file


@register_codec
class HapiCsv(CodecInterface):
    """Codec for HAPI CSV files"""

    def load_variables(self, variables: List[AnyStr], file: Union[Buffer, str, io.IOBase], cache_remote_files=True,
                       **kwargs) -> Optional[Mapping[AnyStr, SpeasyVariable]]:
        hapi_csv_file = load_hapi_csv(file)
        if hapi_csv_file is not None:
            return _hapi_csv_to_speasy_variables(hapi_csv_file, variables)
        return None

    @CacheCall(cache_retention=timedelta(seconds=120), is_pure=True)
    def load_variable(self,
                      variable: AnyStr, file: Union[Buffer, str, io.IOBase],
                      cache_remote_files=True,
                      **kwargs
                      ) -> Optional[SpeasyVariable]:
        return self.load_variables([variable], file, cache_remote_files)[variable]

    def save_variables(self,
                       variables: List[SpeasyVariable],
                       file: Optional[Union[str, io.IOBase]] = None,
                       **kwargs
                       ) -> Union[bool, Buffer]:
        hapi_csv_file = _speasy_variables_to_hapi_csv(variables)
        return save_hapi_csv(hapi_csv_file, file)

    @property
    def supported_extensions(self) -> List[str]:
        return []

    @property
    def supported_mimetypes(self) -> List[str]:
        return []

    @property
    def name(self) -> str:
        return "hapi/csv"



from datetime import timedelta
from functools import wraps
from typing import Callable

from speasy.core.datetime_range import DateTimeRange
from speasy.core.inventory.indexes import ParameterIndex
from speasy.products.variable import merge as var_merge


class SplitLargeRequests(object):
    def __init__(self, threshold: Callable[[ParameterIndex or str], timedelta]):
        self.threshold = threshold

    def __call__(self, get_data: Callable):
        @wraps(get_data)
        def wrapped(wrapped_self, product, start_time, stop_time, **kwargs):
            range = DateTimeRange(start_time, stop_time)
            duration = range.duration
            max_range_per_request = self.threshold(product)
            if duration <= max_range_per_request:
                return get_data(wrapped_self, product=product, start_time=start_time, stop_time=stop_time, **kwargs)
            else:
                fragments = range.split(max_range_per_request)
                return var_merge(
                    [get_data(wrapped_self, product=product, start_time=r.start_time, stop_time=r.stop_time, **kwargs)
                     for r in fragments])

        return wrapped



import numpy
from typing import Callable, Optional, Union, Collection
from speasy.products import SpeasyVariable
import numpy as np
from speasy.core import AnyDateTimeType, make_utc_datetime64, datetime64_to_epoch


def _dt_to_ns(dt: float) -> np.timedelta64:
    return np.timedelta64(int(dt * 1e9), 'ns')


def generate_time_vector(start: AnyDateTimeType, stop: AnyDateTimeType,
                         dt: Union[float, numpy.timedelta64]) -> np.ndarray:
    """Generate a time vector given a start, stop and time step. The time vector will be generated in UTC time zone. The
    time step is in seconds.

    Parameters
    ----------
    start: float or datetime or np.datetime64
        The start time
    stop: float or datetime or np.datetime64
        The stop time
    dt: float or np.timedelta64
        The time step in seconds or as a numpy timedelta64

    Returns
    -------
    np.ndarray
        The time vector as a numpy array of datetime64[ns]
    """
    if type(dt) in (float, int):
        dt = _dt_to_ns(dt)
    return np.arange(make_utc_datetime64(start), make_utc_datetime64(stop), dt, dtype='datetime64[ns]')


class _NumpyInterpolator:
    def __init__(self, x, y):
        self.x = x
        self.y = y

    def __call__(self, new_x):
        return np.interp(new_x, self.x, self.y)


def _interpolate(ref_time: np.ndarray, var: SpeasyVariable, interpolate_callback: Optional[Callable] = None, *args,
                 **kwargs) -> SpeasyVariable:
    res = SpeasyVariable.reserve_like(var, length=len(ref_time))
    res.time[:] = ref_time
    var_epoch = datetime64_to_epoch(var.time)
    res_epoch = datetime64_to_epoch(res.time)
    for col in range(var.values.shape[1]):
        if interpolate_callback is None:
            interpolator = _NumpyInterpolator(var_epoch, var.values[:, col])
        else:
            interpolator = interpolate_callback(var_epoch, var.values[:, col], *args, **kwargs)
        res.values[:, col] = interpolator(res_epoch)
    return res


def resample(var: Union[SpeasyVariable, Collection[SpeasyVariable]], new_dt: Union[float, np.timedelta64],
             interpolate_callback: Optional[Callable] = None,
             *args, **kwargs) -> Union[SpeasyVariable, Collection[SpeasyVariable]]:
    """Resample a variable(s) to a new time step. The time vector will be generated from the start and stop times of the
    input variable. Uses :func:`numpy.interp` to do the resampling by default.

    Parameters
    ----------
    var: SpeasyVariable or Collection[SpeasyVariable]
        The variable(s) to resample
    new_dt: float or np.timedelta64
        The new time step in seconds or as a numpy timedelta64
    interpolate_callback: Callable or None
        The interpolation function to use, defaults to :func:`numpy.interp`

    Returns
    -------
    SpeasyVariable or Collection[SpeasyVariable]
        The resampled variable(s) with all metadata preserved except for the new time axis

    Notes
    -----
    It only supports 1D variables.
    """
    if type(var) in (list, tuple):
        return [resample(v, new_dt, interpolate_callback, *args, **kwargs) for v in var]
    else:
        time = generate_time_vector(var.time[0], var.time[-1] + np.timedelta64(1, 'ns'), new_dt)
        return _interpolate(time, var, interpolate_callback, *args, **kwargs)


def interpolate(ref: Union[np.ndarray, SpeasyVariable], var: Union[SpeasyVariable, Collection[SpeasyVariable]],
                interpolate_callback: Optional[Callable] = None,
                *args, **kwargs) -> Union[SpeasyVariable, Collection[SpeasyVariable]]:
    """Interpolate a variable(s) to a new time vector. The time vector will be taken from the reference variable.
    Uses :func:`numpy.interp` to do the resampling by default.

    Parameters
    ----------
    ref: np.ndarray or SpeasyVariable
        The reference time vector
    var: SpeasyVariable or Collection[SpeasyVariable]
        The variable(s) to interpolate
    interpolate_callback: Callable or None
        The interpolation function to use, defaults to :func:`numpy.interp` (Optional)

    Returns
    -------
    SpeasyVariable or Collection[SpeasyVariable]
        The interpolated variable(s) with all metadata preserved except for the new time axis

    Notes
    -----
    It only supports 1D variables.
    """
    if isinstance(ref, SpeasyVariable):
        ref_time = ref.time
    elif isinstance(ref, np.ndarray) and ref.dtype == np.dtype('datetime64[ns]'):
        ref_time = ref
    else:
        raise ValueError("Invalid reference time vector, must be a numpy array of datetime64[ns] or a SpeasyVariable.")
    if type(var) in (list, tuple):
        return [_interpolate(ref_time, v, interpolate_callback, *args, **kwargs) for v in var]
    return _interpolate(ref_time, var, interpolate_callback, *args, **kwargs)



"""
.. testsetup:: *

   from speasy.core.direct_archive_downloader.direct_archive_downloader import *
"""
import re
from datetime import timedelta, datetime
from functools import partial
import random
from typing import Optional, List, Callable

from dateutil.relativedelta import relativedelta

from speasy.core import make_utc_datetime, AnyDateTimeType
from speasy.core.cache import CacheCall
from speasy.core.any_files import list_files
from speasy.core.codecs import get_codec
from speasy.core.span_utils import intersects
from speasy.products import SpeasyVariable
from speasy.products.variable import merge

# Change to this when we drop Python 3.8
# FileLoaderCallable = Callable[[Optional[str], str, ...], Optional[SpeasyVariable]]
FileLoaderCallable = Callable[..., Optional[SpeasyVariable]]


def apply_date_format(txt: str, date: datetime) -> str:
    if date.hour // 12:
        p = 'PM'
    else:
        p = 'AM'
    return txt.format(Y=date.year, y=str(date.year)[-2:], M=date.month, D=date.day, H=date.hour,
                      j=date.timetuple().tm_yday, I=date.hour % 12,
                      p=p)


@CacheCall(cache_retention=timedelta(hours=12), is_pure=True)
def _read_cdf(url: Optional[str], variable: str, master_cdf_url: Optional[str] = None) -> Optional[SpeasyVariable]:
    if url is None:
        return None
    return get_codec('application/x-cdf').load_variable(file=url, variable=variable, master_cdf_url=master_cdf_url, cache_remote_files=True)


"""
The rationale behind the following function is to randomize the order of execution so we minimize the requests collisions and maximize the throughput.
"""
def randomized_map(f, l):
    indexed_list = list(enumerate(l))
    random.shuffle(indexed_list)
    result = sorted([(i, f(e)) for i, e in indexed_list], key=lambda x: x[0])
    return [e for i, e in result]


def _build_url(url_pattern: str, date: datetime, use_file_list=False) -> Optional[str]:
    base_ulr = apply_date_format(url_pattern, date)
    if not use_file_list:
        return base_ulr
    folder_url, rx = base_ulr.rsplit('/', 1)
    files = sorted(list_files(folder_url, re.compile(rx)))
    if len(files):
        return '/'.join((folder_url, files[-1]))
    return None


def spilt_range(split_frequency: str, start_time: AnyDateTimeType, stop_time: AnyDateTimeType):
    """Given a split frequency (daily, yearly) and a time range, generate the list of start time of each fragment given
    a split frequency (daily, yearly) and a time range to split.

    Parameters
    ----------
    split_frequency : str
        Fragments spilt frequency (daily, monthly, yearly)
    start_time : AnyDateTimeType
        Time range start
    stop_time : AnyDateTimeType
        Time range stop

    Returns
    -------
    List[datetime]
        Ordered list of start time of each fragment composing the given input range

    Examples
    --------
    >>> spilt_range('daily', "2018-01-02", "2018-01-03")
    [datetime.datetime(2018, 1, 2, 0, 0, tzinfo=datetime.timezone.utc), datetime.datetime(2018, 1, 3, 0, 0, tzinfo=datetime.timezone.utc)]
    """
    start: datetime = make_utc_datetime(start_time)
    stop: datetime = make_utc_datetime(stop_time)
    if split_frequency.lower() == "daily":
        start = start.replace(hour=0, minute=0, second=0, microsecond=0)
        return [start + timedelta(days=d) for d in range((stop - start).days + 1)]
    if split_frequency.lower() == "monthly":
        start = start.replace(day=1, hour=0, minute=0, second=0, microsecond=0)
        return [start + relativedelta(months=m) for m in range(relativedelta(stop, start).months + 1)]
    if split_frequency.lower() == "yearly":
        start = start.replace(month=1, day=1, hour=0, minute=0, second=0, microsecond=0)
        return [start + relativedelta(years=y) for y in range(relativedelta(stop, start).years + 1)]
    if split_frequency.lower() == "none":
        return [start]
    raise ValueError(f"Unknown/unimplemented split_frequency: {split_frequency}")


def _parse_date(date: str or datetime, date_format: Optional[str] = None) -> datetime:
    if isinstance(date, datetime) or date_format is None:
        return make_utc_datetime(date)
    if date_format is not None:
        return make_utc_datetime(datetime.strptime(date, date_format))


class RandomSplitDirectDownload:

    @staticmethod
    def overlaps_range(range_start, range_stop, start, stop, version=None):
        start = make_utc_datetime(start)
        stop = make_utc_datetime(stop)
        if start == stop:
            return range_start <= start and range_stop >= stop
        return intersects((start, stop), (range_start, range_stop))

    @staticmethod
    def list_files(split_frequency, url_pattern: str, start_time: AnyDateTimeType, stop_time: AnyDateTimeType,
                   fname_regex: str, date_format=None):

        keep = []
        start_time = make_utc_datetime(start_time)
        stop_time = make_utc_datetime(stop_time)
        for start in spilt_range(split_frequency, start_time, stop_time):

            base_ulr = apply_date_format(url_pattern, start)
            folder_url, rx = base_ulr.rsplit('/', 1)
            files: List[re.Match] = list(
                filter(lambda m: m is not None, map(re.compile(fname_regex).match,
                                                    list_files(folder_url,
                                                               re.compile(rx)))))
            if len(files):
                for index, file in enumerate(files[:-1]):
                    d = file.groupdict()
                    if RandomSplitDirectDownload.overlaps_range(range_start=start_time, range_stop=stop_time,
                                                                start=_parse_date(d['start'], date_format),
                                                                stop=_parse_date(d.get('stop',
                                                                                       files[index + 1].groupdict()[
                                                                                           'start']), date_format)):
                        keep.append(f'{folder_url}/{file.string}')

                d = files[-1].groupdict()
                if RandomSplitDirectDownload.overlaps_range(range_start=start_time, range_stop=stop_time,
                                                            start=_parse_date(d['start'], date_format),
                                                            stop=_parse_date(d.get('stop', max(stop_time,
                                                                                               _parse_date(d['start'],
                                                                                                           date_format))),
                                                                             date_format)):
                    keep.append(f'{folder_url}/{files[-1].string}')
        return keep

    @staticmethod
    def get_product(url_pattern: str, variable: str, start_time: AnyDateTimeType, stop_time: AnyDateTimeType,
                    fname_regex: str, split_frequency: str = "daily", date_format=None,
                    file_reader: FileLoaderCallable = _read_cdf, **kwargs) -> Optional[SpeasyVariable]:
        v = merge(
            randomized_map(partial(file_reader, variable=variable, **kwargs),
                           RandomSplitDirectDownload.list_files(split_frequency=split_frequency,
                                                                url_pattern=url_pattern,
                                                                start_time=start_time, stop_time=stop_time,
                                                                fname_regex=fname_regex,
                                                                date_format=date_format)))
        if v is not None:
            return v[make_utc_datetime(start_time):make_utc_datetime(stop_time)]
        return None


class RegularSplitDirectDownload:

    @staticmethod
    def get_product(url_pattern: str, variable: str, start_time: AnyDateTimeType,
                    stop_time: AnyDateTimeType, use_file_list: bool = False, split_frequency: str = "daily",
                    file_reader: FileLoaderCallable = _read_cdf,
                    **kwargs) -> \
        Optional[SpeasyVariable]:
        v = merge(randomized_map(
            lambda date: file_reader(_build_url(url_pattern, date, use_file_list=use_file_list), variable=variable,
                                     **kwargs),
            spilt_range(split_frequency=split_frequency, start_time=start_time, stop_time=stop_time)))
        if v is not None:
            return v[make_utc_datetime(start_time):make_utc_datetime(stop_time)]
        return None


def get_product(url_pattern: str, split_rule: str, variable: str, start_time: AnyDateTimeType,
                stop_time: AnyDateTimeType, use_file_list: bool = False, file_reader: FileLoaderCallable = _read_cdf,
                codec: Optional[str] = None,
                **kwargs) -> Optional[SpeasyVariable]:
    if codec is not None:
        file_reader = get_codec(codec).load_variable
    if split_rule.lower() == "regular":
        return RegularSplitDirectDownload.get_product(url_pattern, variable, start_time, stop_time,
                                                      use_file_list, file_reader=file_reader, **kwargs)
    if split_rule.lower() == "random":
        return RandomSplitDirectDownload.get_product(url_pattern, variable, start_time, stop_time,
                                                     file_reader=file_reader, **kwargs)

    return None



import os
import sys
from datetime import datetime
from typing import Iterable, List, Optional, Tuple, Union, overload

import numpy as np
import logging

from .. import is_collection, progress_bar
from ..datetime_range import DateTimeRange
from ..inventory.indexes import (CatalogIndex, ComponentIndex,
                                 DatasetIndex, ParameterIndex,
                                 SpeasyIndex, TimetableIndex)
from ...config import core as core_cfg, amda as amda_cfg
from ...products import *
from ...data_providers import (AmdaWebservice, CdaWebservice, CsaWebservice,
                               SscWebservice, GenericArchive)
from ..http import is_server_up

log = logging.getLogger(__name__)

TimeT = Union[str, datetime, float, np.datetime64]
TimeRangeT = Union[DateTimeRange, Tuple[TimeT, TimeT]]
TimeSerieIndexT = Union[ParameterIndex, ComponentIndex]
TimeRangeCollectionT = Union[TimetableIndex, CatalogIndex, Iterable[Iterable[Union[TimeT]]]]

PROVIDERS = {}
amda = None
csa = None
cda = None
ssc = None
archive = None


def init_amda(ignore_disabled_status=False):
    global amda
    if ignore_disabled_status or 'amda' not in core_cfg.disabled_providers():
        if AmdaWebservice.is_server_up():
            amda = AmdaWebservice()
            sys.modules[__name__].amda = amda
            PROVIDERS['amda'] = amda
        else:
            log.warning(f"AMDA server {amda_cfg.entry_point()} is down, disabling AMDA provider")


def init_csa(ignore_disabled_status=False):
    global csa
    if ignore_disabled_status or 'csa' not in core_cfg.disabled_providers():
        if is_server_up(CsaWebservice.BASE_URL):
            csa = CsaWebservice()
            sys.modules[__name__].csa = csa
            PROVIDERS['csa'] = csa
        else:
            log.warning(f"CSA server {CsaWebservice.BASE_URL} is down, disabling CSA provider")


def init_cdaweb(ignore_disabled_status=False):
    global cda
    if ignore_disabled_status or not core_cfg.disabled_providers().intersection({'cda', 'cdaweb'}):
        if is_server_up(CdaWebservice.BASE_URL):
            cda = CdaWebservice()
            sys.modules[__name__].cda = cda
            PROVIDERS['cda'] = cda
            PROVIDERS['cdaweb'] = cda
        else:
            log.warning(f"CDA server {CdaWebservice.BASE_URL} is down, disabling CDA provider")


def init_sscweb(ignore_disabled_status=False):
    global ssc
    if ignore_disabled_status or not core_cfg.disabled_providers().intersection({'ssc', 'sscweb'}):
        if is_server_up(SscWebservice.BASE_URL):
            ssc = SscWebservice()
            sys.modules[__name__].ssc = ssc
            PROVIDERS['ssc'] = ssc
            PROVIDERS['sscweb'] = ssc
        else:
            log.warning(f"SSC server {SscWebservice.BASE_URL} is down, disabling SSC provider")


def init_archive(ignore_disabled_status=False):
    global archive
    if ignore_disabled_status or not core_cfg.disabled_providers().intersection({'archive', 'generic_archive'}):
        archive = GenericArchive()
        sys.modules[__name__].archive = archive
        PROVIDERS['archive'] = archive
        PROVIDERS['generic_archive'] = archive


def init_providers(ignore_disabled_status=False):
    init_amda(ignore_disabled_status=ignore_disabled_status)
    init_csa(ignore_disabled_status=ignore_disabled_status)
    init_cdaweb(ignore_disabled_status=ignore_disabled_status)
    init_sscweb(ignore_disabled_status=ignore_disabled_status)
    init_archive(ignore_disabled_status=ignore_disabled_status)


if 'SPEASY_SKIP_INIT_PROVIDERS' not in os.environ:
    init_providers()


def list_providers() -> List[str]:
    return list(PROVIDERS.keys())


@overload
def get_data(product: CatalogIndex, **kwargs) -> Catalog:
    ...


@overload
def get_data(product: TimetableIndex, **kwargs) -> TimeTable:
    ...


@overload
def get_data(product: str, **kwargs) -> TimeTable or Catalog:
    ...


@overload
def get_data(product: DatasetIndex, start_time: TimeT, stop_time: TimeT, **kwargs) -> Dataset or None:
    ...


@overload
def get_data(product: DatasetIndex, time_range: TimeRangeT, **kwargs) -> Dataset or None:
    ...


@overload
def get_data(product: DatasetIndex, time_range: Iterable[TimeRangeT], **kwargs) -> List[Optional[Dataset]] or None:
    ...


@overload
def get_data(product: TimeSerieIndexT, start_time: TimeT, stop_time: TimeT, **kwargs) -> SpeasyVariable or None:
    ...


@overload
def get_data(product: TimeSerieIndexT, time_range: TimeRangeT, **kwargs) -> SpeasyVariable or None:
    ...


@overload
def get_data(product: TimeSerieIndexT, time_range: Iterable[TimeRangeT], **kwargs) -> List[Optional[
    SpeasyVariable]] or None:
    ...


@overload
def get_data(product: TimeSerieIndexT, time_range: TimeRangeCollectionT, **kwargs) -> List[Optional[
    SpeasyVariable]] or None:
    ...


@overload
def get_data(product: Iterable[TimeSerieIndexT], start_time: TimeT, stop_time: TimeT,
             **kwargs) -> List[Optional[SpeasyVariable]] or None:
    ...


@overload
def get_data(product: Iterable[TimeSerieIndexT], time_range: TimeRangeCollectionT, **kwargs) -> List[List[
    Optional[SpeasyVariable]]] or None:
    ...


def _could_be_datetime(value):
    return type(value) in (str, datetime, np.datetime64, float)


def provider_and_product(path_or_product: str or SpeasyIndex) -> (str, str):
    """Breaks given product in two parts: provider and product UID

    Parameters
    ----------
    path_or_product : str or SpeasyIndex
        The product as SpeasyIndex or path as string you want to split
    Returns
    -------
    (str, str)
        the provider UID and the product UID
    """
    if isinstance(path_or_product, SpeasyIndex):
        return path_or_product.spz_provider().lower(), path_or_product.spz_uid()
    elif type(path_or_product) is str:
        if '/' in path_or_product:
            provider_uid, product_uid = path_or_product.split('/', 1)
            return provider_uid, product_uid
        else:
            raise ValueError(f"Given string does not look like a path {path_or_product}")
    raise TypeError(
        f"Wrong type for {path_or_product}, expecting a string or a SpeasyIndex, got {type(path_or_product)}")


def _scalar_get_data(index, *args, **kwargs):
    provider_uid, product_uid = provider_and_product(index)
    if provider_uid in PROVIDERS:
        return PROVIDERS[provider_uid].get_data(product_uid, *args, **kwargs)
    raise ValueError(f"Can't find a provider for {index}")


def _get_catalog_or_timetable(index, **kwargs):
    return _scalar_get_data(index, **kwargs)


def _get_timeserie1(index, dtrange, **kwargs):
    return _scalar_get_data(index, dtrange[0], dtrange[1], **kwargs)


def _get_timeserie2(index, start, stop, **kwargs):
    return _scalar_get_data(index, start, stop, **kwargs)


def _compile_args(*args, **kwargs):
    if len(args) == 0:
        if 'product' in kwargs:
            args = [kwargs.pop('product')]
    if len(args) == 1:
        if 'start_time' in kwargs and 'stop_time' in kwargs:
            args += [kwargs.pop('start_time'), kwargs.pop('stop_time')]
        elif 'time_range' in kwargs:
            args += [kwargs.pop('time_range')]
    return args, kwargs


def _is_dtrange(value):
    return type(value) is DateTimeRange or (
        hasattr(value, '__len__') and len(value) == 2 and _could_be_datetime(value[0]) and _could_be_datetime(value[1]))


def get_data(*args, **kwargs) -> MaybeAnyProduct:
    """Retrieve requested product(s).
    Speasy gives access to two kind of products, time-dependent products such as physical measurements or trajectories
    and products such as timetables or event catalogs. So depending on which product you want to retrieve :func:`speasy.get_data`
    accepts different sets of arguments:

    - For time-independent products:

    .. highlight:: python
    .. code-block:: python

        get_data(product_or_products, **kwargs)

    - For time-dependent products:

    .. highlight:: python
    .. code-block:: python

        get_data(product_or_products, start_time, stop_time, **kwargs)
        get_data(product_or_products, datetime_range_or_datetime_range, **kwargs)

    Since get_data accepts both at the same time a list of products and a list of ranges, it will always iterate first
    on products then on datetime ranges. In other words, all products will be retrieved for all given datetime ranges.

    Parameters
    ----------
    args :
        Either a time independent or a list of time independent indexes or any combination of time dependent or list of
        time dependent indexes plus a datetime range or a list of datetime ranges. See examples below for more details.
    kwargs :
        For webservice specific keyword arguments check :doc:`/user/data_providers`.

        - disable_proxy: bool
            ignore proxy configuration and always bypass proxy server when True (default: False).
        - disable_cache: bool
            ignore cache content when True (default: False).
        - progress: bool
            show progress bar when True (default: False).

    Returns
    -------
        requested product(s) according to given parameters, either a single product or a collection of products.

    Examples
    --------

    - A simple parameter request on a single tile range:

    >>> import speasy as spz
    >>> spz.get_data("amda/imf_gsm", "2016-10-10", "2016-10-11")
    <speasy.products.variable.SpeasyVariable object at ...>


    - Same with a catalog using Speasy dynamic inventory:

    >>> import speasy as spz
    >>> amda_catalogs = spz.inventories.tree.amda.Catalogs
    >>> spz.get_data(amda_catalogs.SharedCatalogs.EARTH.model_regions_plasmas_cluster_2005)
    <Catalog: model_regions_plasmas_cluster_2005>

    - You can also request a collection of catalogs:

    >>> import speasy as spz
    >>> amda_catalogs = spz.inventories.tree.amda.Catalogs
    >>> spz.get_data([amda_catalogs.SharedCatalogs.EARTH.model_regions_plasmas_cluster_2005,
    ...               amda_catalogs.SharedCatalogs.EARTH.model_regions_plasmas_mms_2019])
    [<Catalog: model_regions_plasmas_cluster_2005>, <Catalog: model_regions_plasmas_mms_2019>]

    - You can also request a parameter for several intervals:

    >>> import speasy as spz
    >>> spz.get_data("amda/imf_gsm", [["2016-10-10", "2016-10-11"],
    ...                               ["2017-10-10", "2017-10-11"]])
    [<speasy.products.variable.SpeasyVariable object at ...>, <speasy.products.variable.SpeasyVariable object at ...>]

    - Several products on a single interval:

    >>> import speasy as spz
    >>> spz.get_data(["amda/imf_gsm", spz.inventories.tree.ssc.Trajectories.wind],
    ...              "2016-10-10", "2016-10-11")
    [<speasy.products.variable.SpeasyVariable object at ...>, <speasy.products.variable.SpeasyVariable object at ...>]

    - Several products for several time ranges:

    >>> import speasy as spz
    >>> data= spz.get_data(["amda/imf_gsm",
    ...                  spz.inventories.tree.ssc.Trajectories.wind],
    ...              [["2016-10-10", "2016-10-11"],
    ...               ["2017-10-10", "2017-10-11"]])
    >>> len(data), len(data[0])
    (2, 2)

    - A catalog or a timetable can also be used as time ranges collection to download a product:

    >>> import speasy as spz
    >>> amda_shared_tt = spz.inventories.tree.amda.TimeTables.SharedTimeTables
    >>> mex_inventory = spz.inventories.tree.amda.Parameters.MEX
    >>> mgs_inventory = spz.inventories.tree.amda.Parameters.MGS
    >>> conj_mex_mgs = spz.get_data(amda_shared_tt.MARS.conjonctions_mex_mgs_2004_0)
    >>> data = spz.get_data(
    ...                     [mex_inventory.ELS.mex_els_all.mex_els_spec,
    ...                      mgs_inventory.MAG.mgs_mag_mso.b_mgs_mso],
    ...                     conj_mex_mgs)
    >>> len(data), len(data[0])
    (2, 28)

    - Can even pass a CatalogIndex or a TimeTableIndex directly:

    >>> import speasy as spz
    >>> amda_shared_tt = spz.inventories.tree.amda.TimeTables.SharedTimeTables
    >>> mex_inventory = spz.inventories.tree.amda.Parameters.MEX
    >>> mgs_inventory = spz.inventories.tree.amda.Parameters.MGS
    >>> data = spz.get_data(
    ...                     [mex_inventory.ELS.mex_els_all.mex_els_spec,
    ...                      mgs_inventory.MAG.mgs_mag_mso.b_mgs_mso],
    ...                     amda_shared_tt.MARS.conjonctions_mex_mgs_2004_0)
    >>> len(data), len(data[0])
    (2, 28)

    """
    args, kwargs = _compile_args(*args, **kwargs)
    if len(args) == 0:
        raise ValueError("You must at least provide a product to retrieve")

    product = args[0]
    if is_collection(product) and not isinstance(product, SpeasyIndex):
        return list(map(lambda p: get_data(p, *args[1:], **kwargs), progress_bar(leave=True, **kwargs)(product)))

    if len(args) == 1:
        return _get_catalog_or_timetable(*args, **kwargs)
    if len(args) == 2:
        t_range = args[1]
        if _is_dtrange(t_range):
            return _get_timeserie1(*args, **kwargs)
        if is_collection(t_range):
            return list(
                map(lambda r: get_data(product, r, *args[2:], **kwargs),
                    progress_bar(leave=False, **kwargs)(t_range)))
        return get_data(product, get_data(t_range), *args[2:], **kwargs)
    if len(args) == 3:
        return _get_timeserie2(*args, **kwargs)



from typing import List, AnyStr, Optional, Mapping, Union
import io
import re
import logging

from datetime import timedelta
import numpy as np

import pyistp
import pycdfpp

from speasy.core.codecs import CodecInterface, register_codec, Buffer
from speasy.core.any_files import any_loc_open
from speasy.core.url_utils import urlparse, is_local_file
from speasy.core.cache import CacheCall
from speasy.products import SpeasyVariable, VariableAxis, VariableTimeAxis, DataContainer

log = logging.getLogger(__name__)
_PTR_rx = re.compile(r".*_PTR(_\d+)?")


def _fix_value_type(value):
    if type(value) in (str, int, float):
        return value
    if type(value) is list:
        return [_fix_value_type(sub_v) for sub_v in value]
    if type(value) is bytes:
        return value.decode('utf-8')
    return str(value)


def _fix_attributes_types(attributes: dict):
    cleaned = {}
    for key, value in attributes.items():
        cleaned[key] = _fix_value_type(value)
    return cleaned


def _is_time_dependent(axis, time_axis_name):
    if axis.attributes.get('DEPEND_TIME', '') == time_axis_name:
        return not axis.is_nrv
    if axis.attributes.get('DEPEND_0', '') == time_axis_name:
        return not axis.is_nrv
    return False


def _make_axis(axis, time_axis_name):
    return VariableAxis(values=axis.values.copy(), meta=_fix_attributes_types(axis.attributes), name=axis.name,
                        is_time_dependent=_is_time_dependent(axis, time_axis_name))


def _build_labels(variable: pyistp.loader.DataVariable):
    if len(variable.values.shape) != 2:
        return _fix_value_type(variable.labels)
    if type(variable.labels) is list and len(variable.labels) == variable.values.shape[1]:
        return _fix_value_type(variable.labels)
    if type(variable.labels) is list and len(variable.labels) == 1:
        return [f"{variable.labels[0]}[{i}]" for i in range(variable.values.shape[1])]
    return [f"component_{i}" for i in range(variable.values.shape[1])]


def _load_variable(istp_loader: pyistp.loader.ISTPLoader, variable) -> SpeasyVariable or None:
    if variable in istp_loader.data_variables():
        var = istp_loader.data_variable(variable)
    elif variable.replace('-', '_') in istp_loader.data_variables():  # THX CSA/ISTP
        var = istp_loader.data_variable(variable.replace('-', '_'))
    else:  # CDA https://cdaweb.gsfc.nasa.gov/WebServices/REST/#Get_Data_GET
        alternative = re.sub(r"[\\/.%!@#^&*()\-+=`~|?<> ]", "$", variable)
        if alternative in istp_loader.data_variables():
            var = istp_loader.data_variable(alternative)
        else:
            return None
    if var is not None:
        time_axis_name = var.axes[0].name
        return SpeasyVariable(
            axes=[VariableTimeAxis(values=var.axes[0].values.copy(),
                                   meta=_fix_attributes_types(var.axes[0].attributes))] + [
                     _make_axis(axis, time_axis_name) for axis in var.axes[1:]],
            values=DataContainer(values=var.values.copy(), meta=_fix_attributes_types(var.attributes),
                                 name=var.name,
                                 is_time_dependent=True),
            columns=_build_labels(var))
    return None


def _load_variables(variables, file=None, buffer=None, master_file=None, master_buffer=None) -> SpeasyVariable or None:
    istp_loader = pyistp.load(file=file, buffer=buffer, master_file=master_file, master_buffer=master_buffer)
    if istp_loader is not None:
        return {variable: _load_variable(istp_loader, variable) for variable in variables}
    return None


def _resolve_url_type(url, prefix="", cache_remote_files=True):
    if url is None:
        return prefix + "file", None
    if type(url) is str:
        if is_local_file(url):
            return prefix + "file", urlparse(url=url).path
        return prefix + "buffer", any_loc_open(url, mode='rb', cache_remote_files=cache_remote_files).read()
    if type(url) in (memoryview, bytes):
        return prefix + "buffer", url
    if hasattr(url, 'read'):
        return prefix + "buffer", url.read()
    return prefix + "file", None


def _simplify_shape(values: np.ndarray) -> np.ndarray:
    if len(values.shape) == 2 and values.shape[1] == 1:
        return np.reshape(values, (-1))
    return values


def _convert_attributes_to_variables(variable_name: str, attrs: Mapping, cdf: pycdfpp.CDF):
    clean_attrs = {}
    for name, attr_v in attrs.items():
        target_name = f"{variable_name}_{name}_{variable_name}"
        if _PTR_rx.match(name):
            cdf.add_variable(
                name=target_name,
                values=attr_v
            )
            clean_attrs[name] = target_name
        else:
            clean_attrs[name] = attr_v
    return clean_attrs


def _write_axis(ax: VariableAxis, cdf: pycdfpp.CDF, compress_variables=False) -> bool:
    data_type = None
    if ax.values.dtype == np.dtype("datetime64[ns]"):
        data_type = pycdfpp.DataType.CDF_TIME_TT2000
    cdf.add_variable(
        name=ax.name,
        values=_simplify_shape(ax.values),
        data_type=data_type,
        compression=pycdfpp.CompressionType.gzip_compression if compress_variables else pycdfpp.CompressionType.no_compression
    )
    return True


def _write_variable(v: SpeasyVariable, cdf: pycdfpp.CDF, already_saved_axes: List[VariableAxis],
                    compress_variables=False) -> bool:
    def _already_in_cdf(ax: VariableAxis):
        for _ax in already_saved_axes:
            if _ax == ax:
                return _ax.name
        return None

    depends = {}
    for index, ax in enumerate(v.axes):
        a = _already_in_cdf(ax)
        if a is None:
            _write_axis(ax, cdf, compress_variables)
            depends[f"DEPEND_{index}"] = ax.name
            already_saved_axes.append(ax)
        else:
            depends[f"DEPEND_{index}"] = a.name
    attributes = v.meta
    attributes.update(depends)
    cdf.add_variable(
        name=v.name,
        values=_simplify_shape(v.values),
        attributes=_convert_attributes_to_variables(variable_name=v.name, attrs=attributes, cdf=cdf),
        compression=pycdfpp.CompressionType.gzip_compression if compress_variables else pycdfpp.CompressionType.no_compression
    )


@register_codec
class IstpCdf(CodecInterface):
    """Codec for ISTP CDF files. This codec is a wrapper around PyISTP library. It supports some variations around the ISTP standard."""

    def load_variables(self,
                       variables: List[AnyStr],
                       file: Union[Buffer, str, io.IOBase],
                       cache_remote_files=True,
                       master_cdf_url: Optional[Union[Buffer, str, io.IOBase]] = None,
                       **kwargs
                       ) -> Optional[Mapping[AnyStr, SpeasyVariable]]:
        kwargs["variables"] = variables
        kwargs.update((_resolve_url_type(file, prefix="", cache_remote_files=cache_remote_files),
                       _resolve_url_type(master_cdf_url, prefix="master_", cache_remote_files=cache_remote_files)))
        return _load_variables(**kwargs)

    @CacheCall(cache_retention=timedelta(seconds=120), is_pure=True)
    def load_variable(self,
                      variable: AnyStr, file: Union[Buffer, str, io.IOBase],
                      cache_remote_files=True,
                      master_cdf_url: Optional[Union[Buffer, str, io.IOBase]] = None,
                      **kwargs
                      ) -> Optional[SpeasyVariable]:
        r = self.load_variables(variables=[variable], file=file, master_cdf_url=master_cdf_url,
                                cache_remote_files=cache_remote_files, **kwargs)
        if r is not None:
            return r.get(variable)
        return None

    def save_variables(self,
                       variables: List[SpeasyVariable],
                       file: Optional[Union[str, io.IOBase]] = None,
                       compress_variables=False,
                       **kwargs
                       ) -> Union[bool, Buffer]:
        cdf = pycdfpp.CDF()
        axes = []
        for variable in variables:
            if not isinstance(variable, SpeasyVariable):
                raise ValueError(f"Expected SpeasyVariable, got {type(variable)}")
            _write_variable(variable, cdf, axes, compress_variables)
        if type(file) is str:
            pycdfpp.save(cdf, file)
            return True
        elif hasattr(file, 'write'):
            file.write(pycdfpp.save(cdf))
            return True
        elif file is None:
            return memoryview(pycdfpp.save(cdf))
        return False

    @property
    def supported_extensions(self) -> List[str]:
        return ["cdf"]

    @property
    def supported_mimetypes(self) -> List[str]:
        return ["application/x-cdf"]

    @property
    def name(self) -> str:
        return self.__class__.__name__



# -*- coding: utf-8 -*-

"""cda package for Space Physics WebServices Client."""

__author__ = """Alexis Jeandet"""
__email__ = 'alexis.jeandet@member.fsf.org'
__version__ = '0.1.0'

import logging
from typing import Optional
from datetime import timedelta

from speasy.config import SPEASY_CONFIG_DIR
from speasy.config import archive as cfg
from speasy.core import AnyDateTimeType, AllowedKwargs
from speasy.core.cdf.inventory_extractor import make_dataset_index
from speasy.core.dataprovider import DataProvider, GET_DATA_ALLOWED_KWARGS
from speasy.core.direct_archive_downloader import get_product
from speasy.core.inventory.indexes import SpeasyIndex, ParameterIndex
from speasy.core.http import is_server_up
from speasy.core.url_utils import host_and_port, is_local_file
from speasy.products.variable import SpeasyVariable
from speasy.core.cache import CacheCall, CACHE_ALLOWED_KWARGS

log = logging.getLogger(__name__)


def _global_inventory_dir():
    import os
    return os.path.join(os.path.dirname(__file__), "../../data/archive")


def user_inventory_dir():
    import os
    return os.path.join(SPEASY_CONFIG_DIR, "archive")


def get_or_make_node(path: str, root: SpeasyIndex) -> SpeasyIndex:
    parts = path.split('/', maxsplit=1)
    name = parts[0]
    if name not in root.__dict__:
        root.__dict__[name] = SpeasyIndex(name=name, provider='archive', uid='')
    if len(parts) == 1:
        return root.__dict__[name]
    return get_or_make_node(parts[1], root.__dict__[name])


@CacheCall(cache_retention=timedelta(seconds=120), is_pure=True)
def _is_up(host, port) -> bool:
    return is_server_up(host=host, port=port)


def _is_reachable(url: str) -> bool:
    host, port = host_and_port(url)
    return _is_up(host, port)


def load_inventory_file(file: str, root: SpeasyIndex):
    import yaml
    with open(file, 'r') as f:
        entries = yaml.safe_load(f)
        for name, entry in entries.items():
            path = f"{entry['inventory_path']}/{name}"
            parent = get_or_make_node(entry['inventory_path'], root)
            entry_meta = {"spz_ga_cfg": entry}
            entry_meta['spz_ga_cfg']['use_file_list'] = entry_meta['spz_ga_cfg'].get('use_file_list', False)
            if is_local_file(entry['master_cdf']) or _is_reachable(entry['master_cdf']):
                dataset = make_dataset_index(entry['master_cdf'], name=name, uid=path, provider='archive',
                                             meta=entry_meta,
                                             params_uid_format=f"{path}/{{var_name}}", params_meta=entry_meta)
                if dataset:
                    parent.__dict__[dataset.spz_name()] = dataset
            else:
                log.warning(f"Master CDF {entry['master_cdf']} is not available, skipping dataset {name}")


class GenericArchive(DataProvider):
    def __init__(self):
        DataProvider.__init__(self, provider_name='archive', provider_alt_names=['generic_archive', 'file'],
                              inventory_disable_proxy=True)

    def build_inventory(self, root: SpeasyIndex):
        from glob import glob
        lookup_dirs = cfg.extra_inventory_lookup_dirs.get()
        lookup_dirs.add(_global_inventory_dir())
        lookup_dirs.add(user_inventory_dir())
        for lookup_dir in lookup_dirs:
            for file in glob(f"{lookup_dir}/*.y*ml"):
                load_inventory_file(file, root)
        return root

    def _parameter_index(self, product: str or ParameterIndex) -> ParameterIndex:
        if type(product) is str:
            if product in self.flat_inventory.parameters:
                return self.flat_inventory.parameters[product]
            else:
                raise ValueError(f"Unknown product {product}")
        elif isinstance(product, ParameterIndex):
            return product
        else:
            raise ValueError(f"Got unexpected type {type(product)}, expecting str or ParameterIndex")

    @AllowedKwargs(GET_DATA_ALLOWED_KWARGS + CACHE_ALLOWED_KWARGS + ['force_refresh'])
    def get_data(self, product: str or ParameterIndex, start_time: AnyDateTimeType, stop_time: AnyDateTimeType,
                 **kwargs) -> Optional[SpeasyVariable]:
        var = self._get_data(product=self._parameter_index(product), start_time=start_time, stop_time=stop_time,
                             **kwargs)
        return var

    def _get_data(self, product: ParameterIndex, start_time: AnyDateTimeType, stop_time: AnyDateTimeType, **kwargs) -> \
        Optional[
            SpeasyVariable]:
        ga_cfg: dict = getattr(product, 'spz_ga_cfg')
        ga_cfg.pop('inventory_path', None)
        ga_cfg.pop('master_cdf', None)
        return get_product(**ga_cfg,
                           variable=product.spz_name(), start_time=start_time, stop_time=stop_time, **kwargs)



from typing import Optional
from .codec_interface import CodecInterface
from speasy.config import core as cfg
import os
import appdirs

__USER_CODECS_DIR__ = f'{appdirs.user_data_dir("speasy", "LPP")}/codecs'

__CODECS__ = {}

"""Speasy codecs registry

This module provides a registry for codecs that can be used to load and save data from different formats. Codecs are registered by their name, supported extensions and supported mimetypes. The registry is used to find the codec that can handle a given file extension or mimetype. Codecs are registered using the `register_codec` decorator.
To register a codec, you must create a class that implements the `CodecInterface` interface and decorate it with the `register_codec` decorator. The codec class will be automatically instantiated and registered in the codecs registry when placed in the user codecs directory. The user codecs directory is defined by the `user_codecs_dir` function.
Alternatively, you can add custom lookup paths to the codecs registry by setting the `user_codecs_extra_dirs` configuration entry to a list of paths. Codecs placed in these paths will be automatically registered in the codecs registry too.
"""


def user_codecs_dir():
    """Get the user codecs directory, any codec placed in this directory will be automatically registered

    Returns
    -------
    str
        The user codecs directory
    """
    return __USER_CODECS_DIR__

def _register_codec(codec: CodecInterface):
    if codec.name in __CODECS__:
        raise ValueError(f"codec {codec.name} already registered")
    __CODECS__[codec.name] = codec
    for ext in codec.supported_extensions:
        if ext in __CODECS__:
            raise ValueError(f"A codec is already registered for extension {ext}")
        __CODECS__[ext] = codec
    for mime in codec.supported_mimetypes:
        if mime in __CODECS__:
            raise ValueError(f"A codec is already registered for mimetype {mime}")
        __CODECS__[mime] = codec


def register_codec(cls):
    """Decorator to register a codec in the codecs registry

    Parameters
    ----------
    cls : CodecInterface
        The codec class to register

    Returns
    -------
    CodecInterface
        The codec class that was registered
    """
    _register_codec(cls())
    return cls


def _list_dir_abs(path: str):
    return [os.path.join(path, f) for f in os.listdir(path) if f.endswith('.py')]

def _load_codec(path: str):
    if path.endswith('.py'):
        exec(open(os.path.join(path, path)).read())


def load_extra_codecs():
    if os.path.exists(user_codecs_dir()):
        list(map(_load_codec, _list_dir_abs(user_codecs_dir())))
    for path in cfg.user_codecs_extra_dirs.get():
        if os.path.exists(path):
            list(map(_load_codec, _list_dir_abs(path)))


load_extra_codecs()


def get_codec(codec: str) -> Optional[CodecInterface]:
    """Get a codec by name, extension or mimetype

    Parameters
    ----------
    codec : str
        Codec name, extension or mimetype

    Returns
    -------
    Optional[CodecInterface]
        The codec that matches the given name, extension or mimetype, or None if no codec is found
    """
    return __CODECS__.get(codec, None)



"""
"""
import json
import logging
import warnings
import os
from datetime import datetime
from typing import Dict, Optional

from ...config import amda as amda_cfg
from ...core import AllowedKwargs, make_utc_datetime, EnsureUTCDateTime
from ...core.http import is_server_up
from ...core.cache import CACHE_ALLOWED_KWARGS, Cacheable, CacheCall
from ...core.dataprovider import (GET_DATA_ALLOWED_KWARGS, ParameterRangeCheck)
from ...core.datetime_range import DateTimeRange
from ...core.inventory.indexes import (CatalogIndex, ParameterIndex, TemplatedParameterIndex,
                                       SpeasyIndex, TimetableIndex, ArgumentIndex)
from ...core.proxy import PROXY_ALLOWED_KWARGS, GetProduct, Proxyfiable, Version
from ...inventories import flat_inventories
from ...products.catalog import Catalog
from ...products.timetable import TimeTable
from ...products.variable import SpeasyVariable

from ...core.impex import ImpexProvider, ImpexEndpoint, to_xmlid
from ...core.impex.exceptions import BadTemplateArgDefinition

log = logging.getLogger(__name__)

amda_provider_name = 'amda'
amda_capabilities = [ImpexEndpoint.AUTH, ImpexEndpoint.OBSTREE, ImpexEndpoint.GETPARAM, ImpexEndpoint.LISTTT,
                     ImpexEndpoint.GETTT, ImpexEndpoint.LISTCAT, ImpexEndpoint.GETCAT, ImpexEndpoint.LISTPARAM,
                     ImpexEndpoint.GETSTATUS]
amda_name_mapping = {
    "dataset": "xmlid",
    "parameter": "xmlid",
    "folder": "name",
    "component": "xmlid",
    "arguments": "name",
    "argument": "key",
    "item": "key"
}

AMDA_MIN_PROXY_VERSION = Version("0.12.0")


def _amda_arguments_to_dict(index):
    if isinstance(index, SpeasyIndex):
        res = {}
        for key, value in index.__dict__.items():
            if isinstance(value, SpeasyIndex) or isinstance(value, str) and key in ['name', 'default', 'type']:
                if index.spz_name() == 'items_list':
                    key = value.key
                res[key] = _amda_arguments_to_dict(value)
        return res
    elif type(index) is not str:
        return str(index)
    return index


def _argument_fits_allowed_values(value:str, argument_desc:ArgumentIndex):
    if argument_desc.type == 'list':
        return value in list(zip(*argument_desc.choices))[1]
    return True

def _stack_level_outside_of_speasy():
    import inspect
    import speasy as spz
    level = 0
    for frame in inspect.stack():
        if os.path.dirname(spz.__file__) not in frame.filename:
            return level
        level += 1
    return level

def _amda_replace_arguments_in_template(product: TemplatedParameterIndex, product_inputs: Dict[str,str]):
    product_id = product.template
    for arg in product.spz_arguments():
        k = arg.key
        v = product_inputs.get(k)
        if v is None:
            v = arg.default
            import speasy as spz
            warnings.warn(f"""Argument {arg.key} is not provided, using default value {v}
You can set Derived Parameters inputs using:
spz.get_data("amda/{product.spz_uid()}, start_time, stop_time, product_inputs={{'{k}': '{arg.default}' }})
""", category=RuntimeWarning, stacklevel=_stack_level_outside_of_speasy())
        if arg.type == 'list' and  not _argument_fits_allowed_values(v, arg):
                raise BadTemplateArgDefinition(f"""Argument {arg.key} has value {v} which is not in the allowed values {arg.choices}""")
        product_id = product_id.replace(f'##{k}##', str(v), 1)
    return product_id


def _amda_get_real_product_id(product_id: str or SpeasyIndex, **kwargs):
    product_id = to_xmlid(product_id)
    product = flat_inventories.__dict__[amda_provider_name].parameters[product_id]
    if isinstance(product, TemplatedParameterIndex) and not hasattr(product, 'predefined'):
        product_inputs = kwargs.get('product_inputs', {})
        real_product_id = _amda_replace_arguments_in_template(product, product_inputs)
    else:
        real_product_id = product_id
    return real_product_id


def _amda_cache_entry_name(prefix: str, product: str, start_time: str, **kwargs):
    output_format: str = kwargs.get('output_format', 'cdf_istp')
    real_product_id = _amda_get_real_product_id(product, **kwargs)
    if output_format.lower() == 'cdf_istp':
        return f"{prefix}/{real_product_id}-cdf_istp/{start_time}"
    else:
        return f"{prefix}/{real_product_id}/{start_time}"


def _amda_get_proxy_parameter_args(start_time: datetime, stop_time: datetime, product: str, **kwargs) -> Dict:
    proxy_args = {'path': f"{amda_provider_name}/{product}", 'start_time': f'{start_time.isoformat()}',
                  'stop_time': f'{stop_time.isoformat()}',
                  'output_format': kwargs.get('output_format', amda_cfg.output_format.get())}
    if kwargs.get('additional_arguments') and isinstance(kwargs.get('additional_arguments'), Dict):
        proxy_args['additional_arguments'] = json.dumps(kwargs.get('additional_arguments'))
    return proxy_args


class AmdaWebservice(ImpexProvider):
    def __init__(self):
        ImpexProvider.__init__(self, provider_name=amda_provider_name, server_url=amda_cfg.entry_point() + "/php/rest",
                               max_chunk_size_days=amda_cfg.max_chunk_size_days(),
                               capabilities=amda_capabilities, name_mapping=amda_name_mapping,
                               username=amda_cfg.username(), password=amda_cfg.password(),
                               output_format=amda_cfg.output_format(), min_proxy_version=AMDA_MIN_PROXY_VERSION)

    @staticmethod
    def is_server_up():
        """Check if AMDA Webservice is up by sending a dummy request to the AMDA Webservice URL with a short timeout.

        Returns
        -------
        bool
            True if AMDA Webservice is up, False otherwise.

        """
        try:
            return is_server_up(url=amda_cfg.entry_point())
        except (Exception,):
            pass
        return False

    def has_time_restriction(self, product_id: str or SpeasyIndex, start_time: str or datetime,
                             stop_time: str or datetime):
        """Check if product is restricted for a given time range.

        Parameters
        ----------
        product_id: str or SpeasyIndex
            product id
        start_time: str or datetime
            desired data start time
        stop_time: str or datetime
            desired data stop time

        Returns
        -------
        bool
            True if product is restricted for the given time range, False otherwise.
        """
        dataset = self.find_parent_dataset(product_id)
        if dataset:
            dataset = self.flat_inventory.datasets[dataset]
            if hasattr(dataset, 'timeRestriction'):
                lower = make_utc_datetime(dataset.timeRestriction)
                upper = make_utc_datetime(dataset.stop_date)
                if lower < upper:
                    return DateTimeRange(lower, upper).intersect(
                        DateTimeRange(start_time, stop_time))
        return False

    def product_version(self, parameter_id: str or ParameterIndex):
        """Get date of last modification of dataset or parameter.

        Parameters
        ----------
        parameter_id: str or ParameterIndex
            parameter id

        Returns
        -------
        str
            product version
        """
        dataset = self.find_parent_dataset(parameter_id)
        if hasattr(self.flat_inventory.datasets[dataset], 'lastModificationDate'):
            return self.flat_inventory.datasets[dataset].lastModificationDate
        return self.flat_inventory.datasets[dataset].lastUpdate

    def get_real_product_id(self, product_id: str or SpeasyIndex, **kwargs):
        return _amda_get_real_product_id(product_id, **kwargs)

    @CacheCall(cache_retention=amda_cfg.user_cache_retention(), is_pure=True)
    def get_timetable(self, timetable_id: str or TimetableIndex, **kwargs) -> Optional[TimeTable]:
        """Get timetable data by ID.

        Parameters
        ----------
        timetable_id: str or TimetableIndex
            time table id

        Returns
        -------
        Optional[TimeTable]
            timetable data

        Examples
        --------

        >>> import speasy as spz
        >>> spz.amda.get_timetable("sharedtimeTable_0")
        <TimeTable: FTE_c1>

        """
        return super().get_timetable(timetable_id, **kwargs)

    @CacheCall(cache_retention=amda_cfg.user_cache_retention(), is_pure=True)
    def get_catalog(self, catalog_id: str or CatalogIndex, **kwargs) -> Optional[Catalog]:
        """Get catalog data by ID.

        Parameters
        ----------
        catalog_id: str or CatalogIndex
            catalog id

        Returns
        -------
        Optional[Catalog]
            catalog data

        Examples
        --------

        >>> import speasy as spz
        >>> spz.amda.get_catalog("sharedcatalog_22")
        <Catalog: model_regions_plasmas_mms_2019>

        """
        return super().get_catalog(catalog_id, **kwargs)

    @CacheCall(cache_retention=amda_cfg.user_cache_retention())
    def get_user_timetable(self, timetable_id: str or TimetableIndex, **kwargs) -> Optional[TimeTable]:
        """Get user timetable. Raises an exception if user is not authenticated.

        Parameters
        ----------
        timetable_id: str or TimetableIndex
            timetable id

        Returns
        -------
        Optional[TimeTable]
            user timetable

        Examples
        --------
        >>> import speasy as spz
        >>> spz.amda.get_user_timetable("tt_0") # doctest: +SKIP
        <TimeTable: test_alexis>

        Warnings
        --------
            Calling :meth:`~speasy.amda.amda.AMDA_Webservice.get_user_timetable` without having defined AMDA_Webservice
            login credentials will result in a :class:`~speasy.core.impex.exceptions.MissingCredentials`
            exception being raised.

        """
        return super().get_user_timetable(timetable_id)

    @CacheCall(cache_retention=amda_cfg.user_cache_retention())
    def get_user_catalog(self, catalog_id: str or CatalogIndex, **kwargs) -> Optional[Catalog]:
        """Get user catalog. Raises an exception if user is not authenticated.


        Parameters
        ----------
        catalog_id: str or CatalogIndex
            catalog id

        Returns
        -------
        Optional[Catalog]
            user catalog

        Examples
        --------

        >>> import speasy as spz
        >>> spz.amda.get_user_catalog("tt_0") # doctest: +SKIP
        <Catalog: test_alexis>

        Warnings
        --------
            Calling :meth:`~speasy.amda.amda.AMDA_Webservice.get_user_catalog` without having defined AMDA_Webservice
            login credentials will result in a :class:`~speasy.core.impex.exceptions.MissingCredentials`
            exception being raised.

        """
        return super().get_user_catalog(catalog_id)

    @AllowedKwargs(
        PROXY_ALLOWED_KWARGS + CACHE_ALLOWED_KWARGS + GET_DATA_ALLOWED_KWARGS +
        ['output_format', 'restricted_period', 'product_inputs'])
    @EnsureUTCDateTime()
    @ParameterRangeCheck()
    @Cacheable(prefix=amda_provider_name, version=product_version, fragment_hours=lambda x: 12,
               entry_name=_amda_cache_entry_name)
    @Proxyfiable(GetProduct, _amda_get_proxy_parameter_args, min_version=AMDA_MIN_PROXY_VERSION)
    def _get_parameter(self, product, start_time, stop_time,
                       extra_http_headers: Dict or None = None, output_format: str or None = None,
                       restricted_period=False, **kwargs) -> \
        Optional[
            SpeasyVariable]:
        """Get parameter data.

        Parameters
        ----------
        product: str or ParameterIndex
            parameter id
        start_time:
            desired data start time
        stop_time:
            desired data stop time
        extra_http_headers: dict
            reserved for internal use
        output_format: str
            request output format in case of success, only CDF_ISTP is supported for now

        Returns
        -------
        Optional[SpeasyVariable]
            product data if available

        Examples
        --------

        >>> import speasy as spz
        >>> import datetime
        >>> imf_data = spz.amda.get_parameter("imf", "2018-01-01", "2018-01-01T01")
        >>> print(imf_data.columns)
        ['imf[0]', 'imf[1]', 'imf[2]']
        >>> print(imf_data.values.shape)
        (225, 3)

        """
        return super()._get_parameter(product, start_time, stop_time, extra_http_headers=extra_http_headers,
                                      output_format=output_format, restricted_period=restricted_period, **kwargs)

    @CacheCall(cache_retention=24 * 60 * 60, is_pure=True)
    def _get_obs_data_tree(self) -> str or None:
        return super()._get_obs_data_tree(add_template_info=True)

    @CacheCall(cache_retention=amda_cfg.user_cache_retention(), is_pure=True)
    def _get_timetables_tree(self) -> str or None:
        return super()._get_timetables_tree()

    @CacheCall(cache_retention=amda_cfg.user_cache_retention(), is_pure=True)
    def _get_user_timetables_tree(self) -> str or None:
        return super()._get_user_timetables_tree()

    @CacheCall(cache_retention=amda_cfg.user_cache_retention(), is_pure=True)
    def _get_catalogs_tree(self) -> str or None:
        return super()._get_catalogs_tree()

    @CacheCall(cache_retention=amda_cfg.user_cache_retention(), is_pure=True)
    def _get_user_catalogs_tree(self) -> str or None:
        return super()._get_user_catalogs_tree()

    @CacheCall(cache_retention=amda_cfg.user_cache_retention(), is_pure=True)
    def _get_derived_parameter_tree(self) -> str or None:
        return super()._get_derived_parameter_tree()



from typing import Protocol, Optional, List, AnyStr, Mapping, Union
import io
from speasy.products import SpeasyVariable


Buffer = Union[memoryview, bytearray, bytes]

class CodecInterface(Protocol):
    """Interface for codecs.

    Codecs are used to load and save data from different formats. Codecs must implement this interface to be registered in the codecs registry.
    """

    def load_variables(self, variables: List[AnyStr], file: Union[Buffer, str, io.IOBase], cache_remote_files=True,
                       **kwargs) -> Optional[Mapping[AnyStr, SpeasyVariable]]:
        """Load variables from a file. The file can be a local file, a remote file or a file-like object.

        Parameters
        ----------
        variables : List[str]
            List of variable names to load
        file : Buffer or str or io.IOBase
            File to load variables from
        cache_remote_files : bool
            Whether to cache remote files
        kwargs
            Additional keyword arguments, codec specific arguments

        Returns
        -------
        Optional[Mapping[str, SpeasyVariable]]
            A dictionary with the variables loaded

        Raises
        ------
        NotImplementedError
            If the method is not implemented

        See Also
        --------
        load_variable, save_variables

        """
        ...

    def load_variable(self, variable: AnyStr, file: Union[Buffer, str, io.IOBase], cache_remote_files=True, **kwargs) -> \
    Optional[SpeasyVariable]:
        """Load a variable from a file. The file can be a local file, a remote file or a file-like object.

        Parameters
        ----------
        variable : str
            Variable name to load
        file : Buffer or str or io.IOBase
            File to load variable from
        cache_remote_files : bool
            Whether to cache remote files
        kwargs
            Additional keyword arguments, codec specific arguments

        Returns
        -------
        Optional[SpeasyVariable]
            The variable loaded

        Raises
        ------
        NotImplementedError
            If the method is not implemented

        See Also
        --------
        load_variables, save_variables

        """
        ...

    def save_variables(self, variables: List[SpeasyVariable], file: Optional[Union[str, io.IOBase]]=None, **kwargs) -> Union[bool,Buffer]:
        """Save variables to a file. The file can be a local file, a remote file or a file-like object.

        Parameters
        ----------
        variables : List[SpeasyVariable]
            List of variables to save
        file : str or io.IOBase
            File to save variables to, if None, return the data as bytes
        kwargs
            Additional keyword arguments, codec specific arguments

        Returns
        -------
        Union[bool, Buffer]
            True if the variables were saved successfully, the data as buffer compatible object if file is None

        Raises
        ------
        NotImplementedError
            If the method is not implemented which is an acceptable behavior for read-only codecs

        See Also
        --------
        load_variables, load_variable

        """
        ...

    @property
    def supported_extensions(self) -> List[str]:
        """List of supported file extensions, without the dot. Do return extensions that could be ambiguous with other codecs."""
        ...

    @property
    def supported_mimetypes(self) -> List[str]:
        """List of supported mime types."""
        ...

    @property
    def name(self) -> str:
        """Codec name. Must be unique."""
        ...



from speasy.core.datetime_range import DateTimeRange
from .base_product import SpeasyProduct
from typing import List
from speasy.core import all_of_type, listify
import pandas as pds


def _all_are_datetime_ranges(dt_list):
    return all_of_type(dt_list, DateTimeRange)


class TimeTable(SpeasyProduct):
    """A TimeTable is basically a collection of DateTimeRange

    """
    __slots__ = ['name', 'meta', '_storage']

    def __init__(self, name: str, meta: dict = None, dt_ranges: List[DateTimeRange] = None):
        super().__init__()
        self.name = name
        self.meta = meta or {}
        self._storage = dt_ranges or []
        if not _all_are_datetime_ranges(self._storage):
            raise TypeError(f"You must provide a list of {DateTimeRange}")

    def __getitem__(self, index):
        return self._storage[index]

    def __len__(self):
        return len(self._storage)

    def __iter__(self):
        return self._storage.__iter__()

    def append(self, dt_range: DateTimeRange or List[DateTimeRange]):
        dt_range = listify(dt_range)
        if not _all_are_datetime_ranges(dt_range):
            raise TypeError(
                f"You must provide a {DateTimeRange} or a List of {DateTimeRange} instead of {type(dt_range)}")
        self._storage += dt_range

    def __iadd__(self, other: DateTimeRange or List[DateTimeRange]):
        self.append(other)
        return self

    def pop(self, index=-1):
        return self._storage.pop(index)

    def to_dataframe(self) -> pds.DataFrame:
        return pds.DataFrame(columns=['start_time', 'stop_time'], data=[(*r,) for r in self])

    def __repr__(self):
        return f"""<TimeTable: {self.name}>"""



from enum import Enum
from typing import Dict, Union
from json.decoder import JSONDecodeError
import logging
import time

from ...core import http
from ...core.http import is_server_up

from .exceptions import MissingCredentials, UnavailableEndpoint

log = logging.getLogger(__name__)


class ImpexEndpoint(Enum):
    """Impex API endpoints.
    """
    AUTH = "auth.php"
    OBSTREE = "getObsDataTree.php"
    LISTTT = "getTimeTablesList.php"
    LISTCAT = "getCatalogsList.php"
    LISTPARAM = "getParameterList.php"
    GETTT = "getTimeTable.php"
    GETCAT = "getCatalog.php"
    GETPARAM = "getParameter.php"
    GETSTATUS = "getStatus.php"
    ISALIVE = "isAlive.php"


class ImpexClient:
    def __init__(self, server_url="", capabilities=None, username="", password="",
                 output_format="CDF_ISTP", time_format='UNIXTIME'):
        self.server_url = server_url
        if capabilities is None:
            capabilities = [ImpexEndpoint.OBSTREE, ImpexEndpoint.GETPARAM]
        self.capabilities = capabilities
        self.username = username
        self.password = password
        self.output_format = output_format
        self.time_format = time_format
        self.use_token = self.is_capable(ImpexEndpoint.AUTH)

    def is_capable(self, api: ImpexEndpoint):
        return api in self.capabilities

    def credential_are_valid(self):
        return self.username != "" and self.password != ""

    def get_credentials(self):
        if self.credential_are_valid():
            return self.username, self.password
        else:
            raise MissingCredentials()

    def reset_credentials(self, username: str = "", password: str = ""):
        self.username = username
        self.password = password

    def reachable(self):
        try:
            return is_server_up(url=f"{self.server_url}/")
        except (Exception,):
            pass
        return False

    def is_alive(self):
        if not self.is_capable(ImpexEndpoint.ISALIVE):
            return True
        return self._send_request(ImpexEndpoint.ISALIVE)

    def auth(self):
        return self._send_request(ImpexEndpoint.AUTH)

    @staticmethod
    def in_progress(result):
        return result == "in progress"

    def get_obs_data_tree(self, use_credentials=False, **kwargs):
        params = {
        }
        if kwargs.get('add_template_info', False):
            params['templateInfo'] = True
        if use_credentials:
            params['userID'], params['password'] = self.get_credentials()
        return self._send_indirect_request(ImpexEndpoint.OBSTREE, params=params)

    def get_time_table_list(self, use_credentials=False):
        params = {}
        if use_credentials:
            params['userID'], params['password'] = self.get_credentials()
        return self._send_indirect_request(ImpexEndpoint.LISTTT, params=params)

    def get_catalog_list(self, use_credentials=False):
        params = {}
        if use_credentials:
            params['userID'], params['password'] = self.get_credentials()
        return self._send_indirect_request(ImpexEndpoint.LISTCAT, params=params)

    def get_derived_parameter_list(self):
        params = {}
        params['userID'], params['password'] = self.get_credentials()
        return self._send_indirect_request(ImpexEndpoint.LISTPARAM, params=params)

    def get_status(self, process_id, extra_http_headers=None):
        params = {
            'id': process_id,
        }
        return self._send_request(ImpexEndpoint.GETSTATUS, params=params, extra_http_headers=extra_http_headers)

    def get_parameter(self, start_time, stop_time, parameter_id, extra_http_headers=None,
                      use_credentials=False, **kwargs):
        params = {
            'startTime': start_time,
            'stopTime': stop_time,
            'parameterID': kwargs.get('real_product_id', parameter_id),
            'outputFormat': kwargs.get('output_format', self.output_format)
        }

        if kwargs.get('time_format'):
            params['timeFormat'] = kwargs.get('time_format')

        if use_credentials:
            params['userID'], params['password'] = self.get_credentials()
        if self.use_token:
            params['token'] = self.auth()
        return self._send_request(ImpexEndpoint.GETPARAM, params=params,
                                  extra_http_headers=extra_http_headers, timeout=5 * 60)

    def get_timetable(self, tt_id, use_credentials=False, **kwargs):
        params = {
            'ttID': tt_id
        }
        if use_credentials:
            params['userID'], params['password'] = self.get_credentials()
        return self._send_request(ImpexEndpoint.GETTT, params=params, **kwargs)

    def get_catalog(self, catalog_id, use_credentials=False, **kwargs):
        params = {
            'catID': catalog_id
        }
        if use_credentials:
            params['userID'], params['password'] = self.get_credentials()
        return self._send_request(ImpexEndpoint.GETCAT, params=params, **kwargs)

    def _request_url(self, endpoint: ImpexEndpoint) -> str:
        if isinstance(endpoint, ImpexEndpoint):
            return f"{self.server_url}/{endpoint.value}"
        else:
            raise TypeError(f"You must provide an {ImpexEndpoint} instead of {type(endpoint)}")

    def _send_indirect_request(self, endpoint: ImpexEndpoint, params: dict = None,
                               timeout: int = http.DEFAULT_TIMEOUT) -> str or None:
        next_url = self._send_request(endpoint=endpoint, params=params, timeout=timeout)
        if '<' in next_url and '>' in next_url:
            next_url = next_url.split(">")[1].split("<")[0]
        r = http.get(next_url, timeout=timeout)
        if r.status_code == 200:
            return r.text.strip()
        return None

    def _send_request(self, endpoint: ImpexEndpoint, params: Dict = None, timeout: int = http.DEFAULT_TIMEOUT,
                      extra_http_headers: Union[Dict, None] = None) -> Union[str, bool, None]:
        if not self.is_capable(endpoint):
            raise UnavailableEndpoint()
        url = self._request_url(endpoint)
        params = params or {}
        http_headers = extra_http_headers or {}
        r = http.get(url, params=params, headers=http_headers, timeout=timeout)
        if r.status_code != 200:
            log.debug(f"Failed: {r.status_code}")
            return None
        try:
            js = r.json()
            if 'success' in js and (js['success'] is True) and ('dataFileURLs' in js):
                log.debug(f"success: {js['dataFileURLs']}")
                return js['dataFileURLs']
            elif "success" in js and (js["success"] is True) and ("status" in js):
                if ImpexClient.in_progress(js['status']):
                    if "id" in js:
                        log.warning("This request duration is too long, consider reducing time range")
                        process_id = js["id"]
                    elif "id" in params:
                        process_id = params["id"]
                    else:
                        log.debug("Cannot retrieve process id")
                        return None
                    default_sleep_time = 10.
                    time.sleep(default_sleep_time)
                    return self.get_status(process_id)
            elif "alive" in js and (js["alive"] is True):
                return True
            else:
                log.debug(f"Failed: {r.text}")
        except JSONDecodeError:
            return r.text.strip()
        return None



from typing import Dict, Optional
import re

_version_regex = re.compile(r"%Q(\.\d+)*")
_date_format_regex = re.compile(r"(%[YymHjdMS])+t?(%[YymHjdMS])+")

substitutions = (
    ('%H', r'[0-2]\d'),
    ('%I', r'[01]\d'),
    ('%p', r'[AP]M'),
    ('%M', r'[0-5]\d'),
    ('%S', r'[0-5]\d')
)

substitutions_yearly = substitutions + (
    ('%Y', '{Y}'),
    ('%y', '{y:02d}'),
    ('%m', r'[01]\d'),
    ('%d', r'[0-3]\d'),
    ('%j', r'[0-3]\d\d')
)

substitutions_monthly = substitutions + (
    ('%Y', '{Y}'),
    ('%y', '{y:02d}'),
    ('%m', '{M:02d}'),
    ('%d', r'[0-3]\d'),
    ('%j', r'[0-3]\d\d')
)

substitutions_daily = substitutions + (
    ('%Y', '{Y}'),
    ('%y', '{y:02d}'),
    ('%m', '{M:02d}'),
    ('%d', '{D:02d}'),
    ('%j', '{j:03d}')
)

none_or_stop_date_substitutions = substitutions + (
    ('%Y', r'[12]\d\d\d'),
    ('%y', r'\d{2}'),
    ('%m', r'[01]\d'),
    ('%d', r'[0-3]\d'),
    ('%j', r'[0-3]\d\d')
)


def _build_date_format(file_naming: str) -> Optional[str]:
    date_format = _date_format_regex.search(file_naming)
    if date_format is None:
        return None
    return date_format.group()


def to_direct_archive_params(file_naming: str, subdivided_by: str, url: str) -> Optional[Dict]:
    if not file_naming.endswith('.cdf'):
        return None

    fname_regex = file_naming

    if subdivided_by == "%Y":
        split_frequency = "yearly"
        subdivided_by = "{Y}"
        for old, new in substitutions_yearly:
            file_naming = file_naming.replace(old, new, 1)
    elif subdivided_by == "%Y/%m":
        split_frequency = "monthly"
        subdivided_by = "{Y}/{M:02d}"
        for old, new in substitutions_monthly:
            file_naming = file_naming.replace(old, new, 1)
    elif subdivided_by == "%Y/%m/%d":
        split_frequency = "daily"
        subdivided_by = "{Y}/{M:02d}/{D:02d}"
        for old, new in substitutions_daily:
            file_naming = file_naming.replace(old, new, 1)
    elif subdivided_by == "None":
        split_frequency = "none"
        subdivided_by = ""
        for old, new in none_or_stop_date_substitutions:
            file_naming = file_naming.replace(old, new, 1)
    else:
        return None

    for old, new in none_or_stop_date_substitutions:
        file_naming = file_naming.replace(old, new, 1)

    file_naming = _version_regex.sub(r".*", file_naming)

    date_format = _build_date_format(fname_regex)
    fname_regex = _date_format_regex.sub(r"(?P<start>\\d+t?T?\\d+)", fname_regex, 1)
    fname_regex = _date_format_regex.sub(r"(?P<stop>\\d+t?T?\\d+)", fname_regex, 1)
    fname_regex = _version_regex.sub(r"(?P<version>.*)", fname_regex)

    return {
        "url_pattern": f"{url}/{subdivided_by}{subdivided_by and '/'}{file_naming}",
        "split_rule": 'random',
        "split_frequency": split_frequency,
        "fname_regex": fname_regex,
        "use_file_list": True,
        "date_format": date_format
    }



import logging
import os.path
from typing import List
import pyistp

from speasy.core.cdf.inventory_extractor import extract_parameters, filter_dataset_meta
from speasy.core.inventory.indexes import ParameterIndex, DatasetIndex, SpeasyIndex

log = logging.getLogger(__name__)


def _patch_parameter(parameter: ParameterIndex, dataset: DatasetIndex):
    parameter.start_date = dataset.start_date
    parameter.stop_date = dataset.stop_date
    parameter.dataset = dataset.spz_uid()
    return parameter


def load_master_cdf(path, dataset: DatasetIndex):
    cdf = pyistp.loader.ISTPLoader(path)
    dataset.__dict__.update(
        {p.spz_name(): p for p in
         map(lambda p: _patch_parameter(p, dataset), extract_parameters(cdf, provider="cda",
                                                                        uid_fmt=f"{dataset.serviceprovider_ID}/{{var_name}}"))})
    dataset.__dict__.update(filter_dataset_meta(cdf))



def _extract_datasets(root: SpeasyIndex) -> List[DatasetIndex]:
    def extract_datasets(node: SpeasyIndex, datasets: List):
        if isinstance(node, DatasetIndex):
            datasets.append(node)
        elif isinstance(node, SpeasyIndex):
            for child in node.__dict__.values():
                extract_datasets(child, datasets)

    datasets = []
    extract_datasets(root, datasets)
    return datasets


def update_tree(root: SpeasyIndex, master_cdf_dir):
    datasets = _extract_datasets(root)
    for dataset in datasets:
        master_cdf_fname = dataset.mastercdf.split('/')[-1]
        full_path = os.path.join(master_cdf_dir, master_cdf_fname)
        if os.path.exists(full_path):
            load_master_cdf(full_path, dataset)



import json
import logging
import platform
import re
import time
from functools import partial
from typing import Optional

import urllib3.response
from urllib3 import PoolManager
from urllib3.util.retry import Retry

from speasy import __version__
from speasy.config import core as core_config
from .url_utils import host_and_port, ApplyRewriteRules

log = logging.getLogger(__name__)

USER_AGENT = f'Speasy/{__version__} {platform.uname()} (SciQLop project)'

DEFAULT_TIMEOUT = 60  # seconds

DEFAULT_DELAY = 5  # seconds

DEFAULT_RETRY_COUNT = 5

STATUS_FORCE_LIST = [500, 502, 504, 413, 429, 503]

RETRY_AFTER_LIST = [429, 503]  # Note: Specific treatment for 429 & 503 error codes (see below)

_HREF_REGEX = re.compile(' href="([A-Za-z0-9.-_]+)">')

pool = PoolManager(num_pools=core_config.urlib_num_pools.get(), maxsize=core_config.urlib_pool_size.get())


class Response:
    def __init__(self, response: urllib3.response.HTTPResponse):
        self._response = response

    @property
    def status_code(self):
        return self._response.status

    @property
    def text(self):
        return self._response.data.decode()

    @property
    def headers(self):
        return self._response.headers

    def json(self):
        return json.loads(self._response.data)

    @property
    def bytes(self):
        return self._response.data

    @property
    def url(self):
        return self._response.geturl()

    @property
    def ok(self):
        return self.status_code in (200, 304)

    def __getattr__(self, item):
        return getattr(self._response, item)

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        return False


class _HttpVerb:
    def __init__(self, verb):
        # cf. https://findwork.dev/blog/advanced-usage-python-requests-timeouts-retries-hooks/
        retry_strategy = Retry(
            total=DEFAULT_RETRY_COUNT,
            backoff_factor=1,
            status_forcelist=STATUS_FORCE_LIST,
            allowed_methods=[verb],
            respect_retry_after_header=True
        )
        # self._adapter = TimeoutHTTPAdapter(max_retries=retry_strategy, timeout=DEFAULT_TIMEOUT)
        # self._http = requests.Session()
        # self._http.mount("https://", self._adapter)
        # self._http.mount("http://", self._adapter)
        self._verb = partial(pool.request, method=verb, retries=retry_strategy)

    @ApplyRewriteRules(is_method=True)
    def __call__(self, url, headers: dict = None, params: dict = None, timeout: int = DEFAULT_TIMEOUT):
        # self._adapter.timeout = timeout
        headers = headers or {}
        headers['User-Agent'] = USER_AGENT
        return Response(self._verb(url=url, headers=headers, fields=params, timeout=timeout))


get = _HttpVerb("GET")
head = _HttpVerb("HEAD")


@ApplyRewriteRules()
def urlopen(url, timeout: int = DEFAULT_TIMEOUT, headers: dict = None):
    headers = {} if headers is None else headers
    headers['User-Agent'] = USER_AGENT
    return Response(pool.urlopen(method="GET", url=url, headers=headers, timeout=timeout))


@ApplyRewriteRules()
def is_server_up(url: Optional[str] = None, host: Optional[str] = None, port: Optional[int] = None, timeout: int = 5,
                 retries=5) -> bool:
    """Checks if a server is up and running. If url is provided, host and port are ignored.

    Parameters
    ----------
    url : Optional[str]
        url to check (scheme://host[:port]), if provided host and port are ignored
    host : Optional[str]
        host to check, if provided port must be provided as well
    port : Optional[int]
        port to check, if provided host must be provided as well
    timeout : int
        timeout in seconds

    Returns
    -------
    bool
        True if server is up and running, False otherwise

    Raises
    ------
    ValueError
        If neither url nor host and port are provided
    """
    if url is not None:
        host, port = host_and_port(url)
    elif host is None or port is None:
        raise ValueError("Either url or host and port must be provided")
    import socket
    try:
        for _ in range(retries):
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.settimeout(timeout)
            result = sock.connect_ex((host, int(port)))
            sock.close()
            if result == 0:
                return True
            log.debug(f"Server {host}:{port} not up yet, retrying in 1 second")
            time.sleep(1.)
        return False
    except Exception as e:
        log.error(f"Error while checking server status: {e}")
        return False



from .codec_interface import CodecInterface, Buffer
from .codecs_registry import register_codec, get_codec, user_codecs_dir

__all__ = ['CodecInterface', 'register_codec', 'get_codec', 'user_codecs_dir']

from . import bundled_codecs
